#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

import gc
import logging
import os
import sys
import textwrap
import traceback
import types
import time
from six import itervalues, iterkeys
from coopr.core import coopr_api

try:
    import yaml
    yaml_available=True
except ImportError:
    yaml_available=False
try:
    import cProfile as profile
except ImportError:
    import profile
try:
    import pstats
    pstats_available=True
except ImportError:
    pstats_available=False

try:
    import IPython
    IPython_available=True
    from IPython.Shell import IPShellEmbed
except:
    IPython_available=False
else:
    ipshell = IPShellEmbed([''],
                banner = '\n# Dropping into Python interpreter',
                exit_msg = '\n# Leaving Interpreter, back to Pyomo\n')

from pyutilib.misc import Options
try:
    from pympler import muppy
    from pympler import summary
    from pympler.asizeof import *
    pympler_available = True
except ImportError:
    pympler_available = False
memory_data = Options()

from coopr.pyomo import *
from coopr.pyomo.base.suffix import active_import_suffix_generator
from coopr.pyomo.base.symbol_map import TextLabeler
from coopr.opt import ProblemFormat
from coopr.opt.base import SolverFactory
from coopr.opt.parallel import SolverManagerFactory
import pyutilib.misc
from coopr.core.plugin import ExtensionPoint, Plugin, implements
from pyutilib.misc import Container
from pyutilib.services import TempfileManager

from coopr.pyomo.expr.linear_repn import linearize_model_expressions

try:
    xrange = xrange
except:
    xrange = range


filter_excepthook=False
modelapi = {    'pyomo_create_model':IPyomoScriptCreateModel,
                'pyomo_create_modeldata':IPyomoScriptCreateModelData,
                'pyomo_print_model':IPyomoScriptPrintModel,
                'pyomo_modify_instance':IPyomoScriptModifyInstance,
                'pyomo_print_instance':IPyomoScriptPrintInstance,
                'pyomo_save_instance':IPyomoScriptSaveInstance,
                'pyomo_print_results':IPyomoScriptPrintResults,
                'pyomo_save_results':IPyomoScriptSaveResults,
                'pyomo_postprocess':IPyomoScriptPostprocess}


logger = logging.getLogger('coopr.pyomo')
start_time = 0.0


@coopr_api(namespace='pyomo.script')
def print_components(data):
    """
    Print information about modeling components supported by Pyomo.
    """
    print("")
    print("----------------------------------------------------------------")
    print("Pyomo Model Components:")
    print("----------------------------------------------------------------")
    components = pyomo.model_components()
    index = pyutilib.misc.sort_index(components)
    for i in index:
        print("")
        print(" "+components[i][0])
        for line in textwrap.wrap(components[i][1], 59):
            print("    "+line)
    print("")
    print("----------------------------------------------------------------")
    print("Pyomo Virtual Sets:")
    print("----------------------------------------------------------------")
    pyomo_sets = pyomo.predefined_sets()
    index = pyutilib.misc.sort_index(pyomo_sets)
    for i in index:
        print("")
        print(" "+pyomo_sets[i][0])
        print("    "+pyomo_sets[i][1])

@coopr_api(namespace='pyomo.script')
def print_solver_help(data):
    """
    Print information about the solvers that are available.
    """
    wrapper = textwrap.TextWrapper(replace_whitespace=False)
    print(wrapper.fill("The --solver option can specify two classes of solvers:  the names of command-line executables that are on the user's path, and predefined solver interfaces.  Command-line executables are assumed to support the 'nl' solver I/O type.  Thus, Pyomo models can be optimized with any solver that employs the AMPL solver library.  The following solver interfaces are recognized by Pyomo:"))
    print("")
    solver_list = SolverFactory.services()
    solver_list = sorted( filter(lambda x: '_' != x[0], solver_list) )
    n = max(map(len, solver_list))
    wrapper = textwrap.TextWrapper(subsequent_indent=' '*(n+9))
    for s in solver_list:
        format = '    %-'+str(n)+'s  %s'
        print(wrapper.fill(format % (s , SolverFactory.doc(s))))
    print("")
    wrapper = textwrap.TextWrapper(subsequent_indent='')
    print(wrapper.fill('These solver interfaces generally support multiple solver I/O types.  The default solver is glpk.'))
    print("")
    print(wrapper.fill('Subsolver options can be specified by with the solver name followed by colon and then the subsolver.  For example, the following specifies that the asl solver will be used:'))
    print('   --asl:PICO')
    print(wrapper.fill('This indicates that the asl solver will launch the PICO executable to perform optimization. Currently, no other solver supports this syntax.'))


@coopr_api(namespace='pyomo.script')
def setup_environment(data):
    """
    Setup Pyomo execution environment
    """
    #
    if data.options.json or not yaml_available:
        data.options.results_format='json'
    else:
        data.options.results_format='yaml'
    #
    global start_time
    start_time = time.time()
    if not data.options.quiet:
        sys.stdout.write('[%8.2f] Setting up Pyomo environment\n' % 0.0)
        sys.stdout.flush()

    #
    # Disable garbage collection
    #
    if data.options.disable_gc:
        gc.disable()
    #
    # Setup management for temporary files
    #
    if not data.options.tempdir is None:
        if not os.path.exists(data.options.tempdir):
            msg =  'Directory for temporary files does not exist: %s'
            raise ValueError(msg % data.options.tempdir)
        TempfileManager.tempdir = data.options.tempdir

    #
    # Configure exception management
    #
    def pyomo_excepthook(etype,value,tb):
        """
        This exception hook gets called when debugging is on. Otherwise,
        run_command in this module is called.
        """
        global filter_excepthook
        if len(data.options.model_file) > 0:
            name = "model " + data.options.model_file
        else:
            name = "model"


        if filter_excepthook:
            action = "loading"
        else:
            action = "running"

        msg = "Unexpected exception (%s) while %s %s:\n" % (etype.__name__, action, name)

        #
        # This handles the case where the error is propagated by a KeyError.
        # KeyError likes to pass raw strings that don't handle newlines
        # (they translate "\n" to "\\n"), as well as tacking on single
        # quotes at either end of the error message. This undoes all that.
        #
        valueStr = str(value)
        if etype == KeyError:
            valueStr = valueStr.replace("\\n","\n")
            if valueStr[0] == valueStr[-1] and valueStr[0] in "\"'":
                valueStr = valueStr[1:-1]

        logger.error(msg+valueStr)

        tb_list = traceback.extract_tb(tb,None)
        i = 0
        if not logger.isEnabledFor(logging.DEBUG) and filter_excepthook:
            while i < len(tb_list):
                if data.options.model_file in tb_list[i][0]:
                    break
                i += 1
            if i == len(tb_list):
                i = 0
        print("\nTraceback (most recent call last):")
        for item in tb_list[i:]:
            print("  File \""+item[0]+"\", line "+str(item[1])+", in "+item[2])
            if item[3] is not None:
                print("    "+item[3])
        sys.exit(1)
    sys.excepthook = pyomo_excepthook


@coopr_api(namespace='pyomo.script')
def apply_preprocessing(data, parser=None):
    """
    Execute preprocessing files

    Required:
        parser: Command line parser object

    Returned:
        error: This is true if an error has occurred.
    """
    #
    if not data.options.quiet:
        sys.stdout.write('[%8.2f] Applying Pyomo preprocessing actions\n' % (time.time()-start_time))
        sys.stdout.flush()
    #
    global filter_excepthook
    #
    #
    # Setup solver and model
    #
    #
    if len(data.options.model_file) == 0:
        parser.print_help()
        data.error = True
        return data
    #
    if not data.options.preprocess is None:
        for file in data.options.preprocess:
            preprocess = pyutilib.misc.import_file(file)
    #
    for ep in ExtensionPoint(IPyomoScriptPreprocess):
        ep.apply( options=data.options )
    #
    # Verify that files exist
    #
    for file in [data.options.model_file]+data.options.data_files:
        if not os.path.exists(file):
            raise IOError("File "+file+" does not exist!")
    #
    filter_excepthook=True
    data.options.usermodel = pyutilib.misc.import_file(data.options.model_file)
    filter_excepthook=False

    usermodel_dir = dir(data.options.usermodel)
    data.options._usermodel_plugins = []
    for key in modelapi:
        if key in usermodel_dir:
            class TMP(Plugin):
                implements(modelapi[key], service=True)
                def __init__(self):
                    self.fn = getattr(data.options.usermodel, key)
                def apply(self,**kwds):
                    return self.fn(**kwds)
            tmp = TMP()
            data.options._usermodel_plugins.append( tmp )
            #print "HERE", modelapi[key], coopr.core.plugin.interface_services[modelapi[key]]

    #print "HERE", data.options._usermodel_plugins

    if 'pyomo_preprocess' in usermodel_dir:
        if data.options.model_name in usermodel_dir:
            msg = "Preprocessing function 'pyomo_preprocess' defined in file" \
                  " '%s', but model is already constructed!"
            raise SystemExit(msg % data.options.model_file)
        getattr(data.options.usermodel, 'pyomo_preprocess')( options=data.options )
    #
    return data

@coopr_api(namespace='pyomo.script')
def create_model(data):
    """
    Create instance of Pyomo model.
    
    Return:
        model:      Model object.
        instance:   Problem instance.
        symbol_map: Symbol map created when writing model to a file.
        filename:    Filename that a model instance was written to.
    """
    #
    if not data.options.quiet:
        sys.stdout.write('[%8.2f] Creating model\n' % (time.time()-start_time))
        sys.stdout.flush()
    #
    if (pympler_available is True) and (not data.options.profile_memory is None) and (data.options.profile_memory >= 1):
        global memory_data
        mem_used = muppy.get_size(muppy.get_objects())
        data.options.max_memory = mem_used
        print("   Total memory = %d bytes prior to model construction" % mem_used)

    #
    # Create Model
    #
    ep = ExtensionPoint(IPyomoScriptCreateModel)
    model_name = 'model'
    if data.options.model_name is not None: model_name = data.options.model_name

    if model_name in dir(data.options.usermodel):
        if len(ep) > 0:
            msg = "Model construction function 'create_model' defined in "    \
                  "file '%s', but model is already constructed!"
            raise SystemExit(msg % data.options.model_file)
        model = getattr(data.options.usermodel, model_name)

        if model is None:
            msg = "'%s' object is 'None' in module %s"
            raise SystemExit(msg % (model_name, data.options.model_file))
            sys.exit(0)

    else:
        if len(ep) == 0:
            msg = "Neither '%s' nor 'pyomo_create_model' are available in "    \
                  'module %s'
            raise SystemExit(msg % ( model_name, data.options.model_file ))
        elif len(ep) > 1:
            msg = 'Multiple model construction plugins have been registered!'
            raise SystemExit(msg)
        else:
            model_options = data.options.model_options
            if model_options is None:
                model_options = []
            model = ep.service().apply( options = pyutilib.misc.Container(*data.options), model_options=pyutilib.misc.Container(*model_options) )
    #
    for ep in ExtensionPoint(IPyomoScriptPrintModel):
        ep.apply( options=data.options, model=model )

    #
    # Disable canonical repn for ASL solvers, and if the user has specified as such (in which case, we assume they know what they are doing!).
    #
    # Likely we need to change the framework so that canonical repn
    # is not assumed to be required by all solvers?
    #
    if not data.options.solver is None and data.options.solver.startswith('asl'):
        model.skip_canonical_repn = True
    elif data.options.skip_canonical_repn is True:
        model.skip_canonical_repn = True

    #
    # Create Problem Instance
    #
    ep = ExtensionPoint(IPyomoScriptCreateModelData)
    if len(ep) > 1:
        msg = 'Multiple model data construction plugins have been registered!'
        raise SystemExit(msg)

    if len(ep) == 1:
        modeldata = ep.service().apply( options=data.options, model=model )
    else:
        modeldata = ModelData()

    if len(data.options.data_files) > 1:
        #
        # Load a list of *.dat files
        #
        for file in data.options.data_files:
            suffix = (file).split(".")[-1]
            if suffix != "dat":
                msg = 'When specifiying multiple data files, they must all '  \
                      'be *.dat files.  File specified: %s'
                raise SystemExit(msg % str( file ))

            modeldata.add(file)

        modeldata.read(model)

        if not data.options.profile_memory is None:
            instance = model.create(modeldata, namespaces=data.options.namespaces, profile_memory=data.options.profile_memory, report_timing=data.options.report_timing)
        else:
            instance = model.create(modeldata, namespaces=data.options.namespaces, report_timing=data.options.report_timing)

    elif len(data.options.data_files) == 1:
        #
        # Load a *.dat file or process a *.py data file
        #
        suffix = (data.options.data_files[0]).split(".")[-1].lower()
        if suffix == "dat":
            if not data.options.profile_memory is None:
                instance = model.create(data.options.data_files[0], namespaces=data.options.namespaces, profile_memory=data.options.profile_memory, report_timing=data.options.report_timing)
            else:
                instance = model.create(data.options.data_files[0], namespaces=data.options.namespaces, report_timing=data.options.report_timing)
        elif suffix == "py":
            userdata = pyutilib.misc.import_file(data.options.data_files[0])
            if "modeldata" in dir(userdata):
                if len(ep) == 1:
                    msg = "Cannot apply 'pyomo_create_modeldata' and use the" \
                          " 'modeldata' object that is provided in the model"
                    raise SystemExit(msg)

                if userdata.modeldata is None:
                    msg = "'modeldata' object is 'None' in module %s"
                    raise SystemExit(msg % str( data.options.data_files[0] ))

                modeldata=userdata.modeldata

            else:
                if len(ep) == 0:
                    msg = "Neither 'modeldata' nor 'pyomo_create_modeldata'"  \
                          'is defined in module %s'
                    raise SystemExit(msg % str( data.options.data_files[0] ))

            modeldata.read(model)
            if not data.options.profile_memory is None:
                instance = model.create(modeldata, namespaces=data.options.namespaces, profile_memory=data.options.profile_memory, report_timing=data.options.report_timing)
            else:
                instance = model.create(modeldata, namespaces=data.options.namespaces, report_timing=data.options.report_timing)
        elif suffix == "yml" or suffix == 'yaml':
            try:
                import yaml
            except:
                msg = "Cannot apply load data from a YAML file: PyYaml is not installed"
                raise SystemExit(msg)

            modeldata = yaml.load(open(data.options.data_files[0]))
            if not data.options.profile_memory is None:
                instance = model.create(modeldata, namespaces=data.options.namespaces, profile_memory=data.options.profile_memory, report_timing=data.options.report_timing)
            else:
                instance = model.create(modeldata, namespaces=data.options.namespaces, report_timing=data.options.report_timing)
        else:
            raise ValueError("Unknown data file type: "+data.options.data_files[0])
    else:
        if not data.options.profile_memory is None:
            instance = model.create(modeldata, namespaces=data.options.namespaces, profile_memory=data.options.profile_memory, report_timing=data.options.report_timing)
        else:
            instance = model.create(modeldata, namespaces=data.options.namespaces, report_timing=data.options.report_timing)

    if data.options.linearize_expressions is True:
        linearize_model_expressions(instance)

    #
    modify_start_time = time.time()
    for ep in ExtensionPoint(IPyomoScriptModifyInstance):
        if data.options.report_timing is True:
            tick = time.time()
        ep.apply( options=data.options, model=model, instance=instance )
        if data.options.report_timing is True:
            print("      %6.2f seconds to apply %s" % (time.time() - tick, type(ep)))
            tick = time.time()
    #
    if data.options.report_timing is True:
        total_time = time.time() - modify_start_time
        print("      %6.2f seconds required for problem transformations" % total_time)
        
    if logger.isEnabledFor(logging.DEBUG):
        print("MODEL INSTANCE")
        instance.pprint()
        print("")

    for ep in ExtensionPoint(IPyomoScriptPrintInstance):
        ep.apply( options=data.options, instance=instance )

    fname=None
    symbol_map=None
    #if options.save_model is None and options.debug:
        #options.save_model = 'unknown.lp'
    if not data.options.save_model is None:

        if data.options.report_timing is True:
            write_start_time = time.time()

        if data.options.save_model == True:
            if data.options.format in (ProblemFormat.cpxlp, ProblemFormat.lpxlp):
                fname = (data.options.data_files[0])[:-3]+'lp'
            else:
                fname = (data.options.data_files[0])[:-3]+str(data.options.format)
            format=data.options.format
        else:
            fname = data.options.save_model
            format=None
        (fname, symbol_map) = instance.write(filename=fname, format=format, io_options={"symbolic_solver_labels" : data.options.symbolic_solver_labels})
        if not data.options.quiet:
            if not os.path.exists(fname):
                print("ERROR: file "+fname+" has not been created!")
            else:
                print("Model written to file '"+str(fname)+"'")

        if data.options.report_timing is True:
            total_time = time.time() - write_start_time
            print("      %6.2f seconds required to write file" % total_time)

        if (pympler_available is True) and (data.options.profile_memory >= 2):
            print("")
            print("      Summary of objects following file output")
            post_file_output_summary = summary.summarize(muppy.get_objects())
            summary.print_(post_file_output_summary, limit=100)

            print("")

    for ep in ExtensionPoint(IPyomoScriptSaveInstance):
        ep.apply( options=data.options, instance=instance )

    if (pympler_available is True) and (data.options.profile_memory >= 1):
        mem_used = muppy.get_size(muppy.get_objects())
        if mem_used > data.options.max_memory:
            data.options.max_memory = mem_used
        print("   Total memory = %d bytes following Pyomo instance creation" % mem_used)

    return pyutilib.misc.Options(
                    model=model, instance=instance,
                    symbol_map=symbol_map, filename=fname )

@coopr_api(namespace='pyomo.script')
def apply_optimizer(data, instance=None):
    """
    Perform optimization with a concrete instance

    Required:
        instance:   Problem instance.

    Returned:
        results:    Optimization results. 
        opt:        Optimizer object.
    """
    #
    if not data.options.quiet:
        sys.stdout.write('[%8.2f] Applying solver\n' % (time.time()-start_time))
        sys.stdout.flush()
    #
    #
    # Create Solver and Perform Optimization
    #
    solver = data.options.solver
    if solver is None:
        raise ValueError("Problem constructing solver:  no solver specified")

    opt = SolverFactory( solver, solver_io=data.options.solver_io )
    if opt is None:
        raise ValueError("Problem constructing solver `%s`" % str(solver))

    opt.keepfiles=data.options.keepfiles or data.options.log

    from coopr.pyomo.base.plugin import registered_callback
    for name in registered_callback:
        opt.set_callback(name, registered_callback[name])

    opt.symbolic_solver_labels = data.options.symbolic_solver_labels

    if data.options.timelimit == 0:
        data.options.timelimit=None

    if not data.options.solver_suffixes is None:
        for suffix_name in data.options.solver_suffixes:
            if suffix_name[0] in ['"',"'"]:
                suffix_name = suffix[1:-1]
            # Don't redeclare the suffix if it already exists
            suffix = getattr(instance,suffix_name,None)
            if suffix is None:
                setattr(instance,suffix_name,Suffix(direction=Suffix.IMPORT))
            else:
                raise ValueError("Problem declaring solver suffix %s. A component "\
                                  "with that name already exists on model %s." % (suffix_name, instance.name))

    if not data.options.solver_options is None:
        opt.set_options(" ".join(data.options.solver_options))

    if data.options.smanager_type is None:
        solver_mngr = SolverManagerFactory( 'serial' )
    else:
        solver_mngr = SolverManagerFactory( data.options.smanager_type )

    if solver_mngr is None:
        msg = "Problem constructing solver manager '%s'"
        raise ValueError(msg % str( data.options.smanager_type ))

    results = solver_mngr.solve( instance, 
                                 opt=opt, 
                                 tee=data.options.tee, 
                                 timelimit=data.options.timelimit )

    if results == None:
        raise ValueError("opt.solve returned None")

    if (pympler_available is True) and (not data.options.profile_memory is None) and (data.options.profile_memory >= 1):
        global memory_data
        mem_used = muppy.get_size(muppy.get_objects())
        if mem_used > data.options.max_memory:
            data.options.max_memory = mem_used
        print("   Total memory = %d bytes following optimization" % mem_used)

    return pyutilib.misc.Options(results=results, opt=opt)


@coopr_api(namespace='pyomo.script')
def process_results(data, instance=None, results=None, opt=None):
    """
    Process optimization results.

    Required:
        instance:   Problem instance.
        results:    Optimization results object.
        opt:        Optimizer object.
    """
    #
    if not data.options.quiet:
        sys.stdout.write('[%8.2f] Processing results\n' % (time.time()-start_time))
        sys.stdout.flush()
    #
    if data.options.log:
        print("")
        print("==========================================================")
        print("Solver Logfile:",opt.log_file)
        print("==========================================================")
        print("")
        INPUT = open(opt.log_file, "r")
        for line in INPUT:
            print(line,)
        INPUT.close()
    #
    # JDS: FIXME: This is a HACK for the ASL.  The SOL file does not
    # actually contain the objective values, so we must calculate them
    # ourselves.  Ideally, this should be done as part of the ASL solver
    # (i.e., part of reading in a SOL file), however, given the current
    # workflow structure, the necessary information is not present
    # (i.e., results reading is supposed to be independent of the
    # instance and the symbol_map).  This should be revisited as part of
    # any workflow overhaul.
    if instance is not None and results is not None and \
           results._symbol_map is not None:
        # We need the symbol map in order to translate the strings
        # coming back in the results object to the actual varvalues in
        # the instance
        _symbolMap = results._symbol_map

        # This is a lot of work to get the flattened list of objectives
        # (especially since all the solvers are single-objective)...
        # But this is safe for multi-objective use (both multiple
        # objectives and indexed objectives)
        _objectives = []
        for obj in itervalues(instance.components(Objective)):
            _objectives.extend(obj.values())
        _nObj = len(_objectives)
        
        labeler = None
        for _result in xrange(len(results.solution)):
            _soln = results.solution[_result] 

            # The solution objective keys may have data on them (like suffixes) 
            # yet lack a value. This is still an ugly hack, but we really only
            # need to go through the rest of this process for those objectives
            # results that lack a .value attribute.
            _incomplete_objectives = []

            for obj in _objectives:
                try:
                    if (_symbolMap.getObject("__default_objective__") is obj) \
                            and ("__default_objective__" in _soln.objective):
                        _soln.objective["__default_objective__"].value
                    else:
                        if labeler is None:
                            labeler = TextLabeler()
                        lbl = _symbolMap.getSymbol(obj, labeler)
                        if lbl in _soln.objective:
                            _soln.objective[lbl].value
                        else:
                            _incomplete_objectives.append(obj)
                except AttributeError:
                    _incomplete_objectives.append(obj)

            if len(_incomplete_objectives) == 0:
                continue

            if labeler is None:
                labeler = TextLabeler()

            # Save the original instance values... that way the original
            # instance does not change "unexpectedly"
            _orig_val_map = {}

            # We need to map the symbols returned by the solver results
            # to their "official" symbols.  This is because the ASL
            # actually returns *aliases* as the names in the results
            # object <sigh>.
            _results_name_map = {}
            for var in iterkeys(_soln.variable):
                # dangerous: this assumes that all results from the solver
                # actually went through the symbol map
                _name = _symbolMap.getSymbol(_symbolMap.getObject(var), labeler)
                _results_name_map[_name] = var

            # Pull the variables out of the objective, override them
            # with the results from the solver, and evaluate each
            # objective
            for obj in _objectives:
                for var in coopr.pyomo.base.expr.identify_variables( obj.expr, False ):
                    # dangerous: this assumes that all variables
                    # actually went through the symbol map
                    s = results._symbol_map.getSymbol(var, labeler)
                    if s not in _orig_val_map:
                        _orig_val_map.setdefault(s, (var, var.value))
                        if s in _results_name_map:
                            var.value = _soln.variable[_results_name_map[s]]['Value']
                        else:
                            var.value = 0.0
                try:
                    obj_val = value(obj.expr)
                except NotImplementedError:
                    obj_val = 'unknown'

                if _symbolMap.getObject("__default_objective__") is obj:
                    _soln.objective["__default_objective__"].value = obj_val
                else:
                    _soln.objective[ _symbolMap.getSymbol(obj, labeler) ].value = obj_val

            # Finally, put the variables back to their original values
            for var, val in itervalues(_orig_val_map):
                var.value = val
    #
    try:
        # transform the results object into human-readable names.
        # IMPT: the resulting object won't be loadable - it's only for output.
        transformed_results = instance.update_results(results)
    except Exception:
        print("Problem updating solver results")
        raise
    #
    if not data.options.show_results:
        if data.options.save_results:
            results_file = data.options.save_results
        elif data.options.results_format == 'yaml':
            results_file = 'results.yml'
        else:
            results_file = 'results.json'
        transformed_results.write(filename=results_file, format=data.options.results_format)
        if not data.options.quiet:
            print("    Number of solutions: "+str(len(transformed_results.solution)))
            if len(transformed_results.solution) > 0:
                print("    Solution Information")
                print("      Gap: "+str(transformed_results.solution[0].gap))
                print("      Status: "+str(transformed_results.solution[0].status))
                if len(transformed_results.solution[0].objective) == 1:
                    key = transformed_results.solution[0].objective.keys()[0]
                    print("      Function Value: "+str(transformed_results.solution[0].objective[key].value))
            print("    Solver results file: "+results_file)
    #
    ep = ExtensionPoint(IPyomoScriptPrintResults)
    if len(ep) == 0:
        try:
            instance.load(results)
        except Exception:
            print("Problem loading solver results")
            raise
    if data.options.show_results:
        print("")
        transformed_results.write(num=1, format=data.options.results_format)
        print("")
    #
    if data.options.summary:
        print("")
        print("==========================================================")
        print("Solution Summary")
        print("==========================================================")
        if len(results.solution(0).variable) > 0:
            print("")
            display(instance)
            print("")
        else:
            print("No solutions reported by solver.")
    #
    for ep in ExtensionPoint(IPyomoScriptPrintResults):
        ep.apply( options=data.options, instance=instance, results=results )
    #
    for ep in ExtensionPoint(IPyomoScriptSaveResults):
        ep.apply( options=data.options, instance=instance, results=results )
    #
    if (pympler_available is True) and (data.options.profile_memory >= 1):
        global memory_data
        mem_used = muppy.get_size(muppy.get_objects())
        if mem_used > data.options.max_memory:
            data.options.max_memory = mem_used
        print("   Total memory = %d bytes following results processing" % mem_used)

@coopr_api(namespace='pyomo.script')
def apply_postprocessing(data, instance=None, results=None):
    """
    Apply post-processing steps.

    Required:
        instance:   Problem instance.
        results:    Optimization results object.
    """
    #
    if not data.options.quiet:
        sys.stdout.write('[%8.2f] Applying Pyomo postprocessing actions\n' % (time.time()-start_time))
        sys.stdout.flush()
    #
    for file in data.options.postprocess:
        postprocess = pyutilib.misc.import_file(file)
        if "pyomo_postprocess" in dir(postprocess):
            postprocess.pyomo_postprocess(data.options, instance,results)
    for ep in ExtensionPoint(IPyomoScriptPostprocess):
        ep.apply( options=data.options, instance=instance, results=results )

    if (pympler_available is True) and (data.options.profile_memory >= 1):
        mem_used = muppy.get_size(muppy.get_objects())
        if mem_used > data.options.max_memory:
            data.options.max_memory = mem_used
        print("   Total memory = %d bytes upon termination" % mem_used)

@coopr_api(namespace='pyomo.script')
def finalize(data, model=None, instance=None, results=None):
    """
    Perform final actions to finish the execution of the pyomo script.

    This function prints statistics related to the execution of the pyomo script.
    Additionally, this function will drop into the python interpreter if the `interactive` 
    option is `True`.

    Required:
        model:      A pyomo model object.

    Optional:
        instance:   A problem instance derived from the model object.
        results:    Optimization results object.
    """
    #
    # Deactivate and delete plugins
    #
    ##import gc
    ##print "HERE - usermodel_plugins"
    ##_tmp = data.options._usermodel_plugins[0]
    cleanup()
    data.options._usermodel_plugins = []
    ##gc.collect()
    ##print gc.get_referrers(_tmp)
    ##import coopr.pyomo.base.plugin
    ##print coopr.core.plugin.interface_services[coopr.pyomo.base.plugin.IPyomoScriptSaveResults]
    ##print "HERE - usermodel_plugins"
    ##
    if not data.options.quiet:
        sys.stdout.write('[%8.2f] Pyomo Finished\n' % (time.time()-start_time))
        if (pympler_available is True) and (data.options.profile_memory >= 1):
            sys.stdout.write('Maximum memory used = %d bytes\n' % data.options.max_memory)
        sys.stdout.flush()
    #
    model=model
    instance=instance
    results=results
    #
    if data.options.interactive:
        if IPython_available:
            ipshell()
        else:
            import code
            shell = code.InteractiveConsole(locals())
            print('\n# Dropping into Python interpreter')
            shell.interact()
            print('\n# Leaving Interpreter, back to Pyomo\n')


@coopr_api(namespace='pyomo.script')
def run_command(command=None, parser=None, args=None, name='unknown', data=None):
    """
    Execute a function that processes command-line arguments and
    then calls a command-line driver.

    This function provides a generic facility for executing a command
    function is rather generic.  This function is segregated from
    the driver to enable profiling of the command-line execution.

    Required:
        command:    The name of a function that will be executed to perform process the command-line
                    options with a parser object.
        parser:     The parser object that is used by the command-line function.

    Optional:
        args:       Command-line arguments that are parsed.  If this value is `None`, then the
                    arguments in `sys.argv` are used to parse the command-line.
        name:       Specifying the name of the command-line (for error messages).
        data:       A container of labeled data.

    Returned:
        retval:     Return values from the command-line execution.
        errorcode:  0 if Pyomo ran successfully
    """
    #
    #
    # Parse command-line options
    #
    #
    retval = None
    errorcode = 0
    try:
        _options = parser.parse_args(args=args)
        # Replace the parser options object with a pyutilib.misc.Options object
        options = pyutilib.misc.Options()
        for key in dir(_options):
            if key[0] != '_':
                val = getattr(_options, key)
                if not isinstance(val, types.MethodType):
                    options[key] = val
    except SystemExit:
        # the parser throws a system exit if "-h" is specified - catch
        # it to exit gracefully.
        return Container(retval=retval, errorcode=errorcode)
    #
    # Configure the logger
    #
    if options.quiet:
        logging.getLogger('coopr.pyomo').setLevel(logging.ERROR)
        logging.getLogger('coopr').setLevel(logging.ERROR)
        logging.getLogger('pyutilib').setLevel(logging.ERROR)        
    if options.warning:
        logging.getLogger('coopr.pyomo').setLevel(logging.WARNING)
        logging.getLogger('coopr').setLevel(logging.WARNING)
        logging.getLogger('pyutilib').setLevel(logging.WARNING)
    if options.info:
        logging.getLogger('coopr.pyomo').setLevel(logging.INFO)
        logging.getLogger('coopr').setLevel(logging.INFO)
        logging.getLogger('pyutilib').setLevel(logging.INFO)
    if options.verbose > 0:
        if options.verbose >= 1:
            logger.setLevel(logging.DEBUG)
        if options.verbose >= 2:
            logging.getLogger('coopr').setLevel(logging.DEBUG)
        if options.verbose >= 3:
            logging.getLogger('pyutilib').setLevel(logging.DEBUG)
    if options.debug:
        logging.getLogger('coopr.pyomo').setLevel(logging.DEBUG)
        logging.getLogger('coopr').setLevel(logging.DEBUG)
        logging.getLogger('pyutilib').setLevel(logging.DEBUG)
    if options.logfile:
        logging.getLogger('coopr.pyomo').handlers = []
        logging.getLogger('coopr').handlers = []
        logging.getLogger('pyutilib').handlers = []
        logging.getLogger('coopr.pyomo').addHandler( logging.FileHandler(options.logfile, 'w'))
        logging.getLogger('coopr').addHandler( logging.FileHandler(options.logfile, 'w'))
        logging.getLogger('pyutilib').addHandler( logging.FileHandler(options.logfile, 'w'))
    #
    # Setup I/O redirect to a file
    #
    logfile = getattr(options, 'output', None)
    if not logfile is None:
        pyutilib.misc.setup_redirect(logfile)
    #
    # Call the main Pyomo runner with profiling
    #
    TempfileManager.push()
    if options.profile > 0:
        if not pstats_available:
            if not logfile is None:
                pyutilib.misc.reset_redirect()
            msg = "Cannot use the 'profile' option.  The Python 'pstats' "    \
                  'package cannot be imported!'
            raise ValueError(msg)
        tfile = TempfileManager.create_tempfile(suffix=".profile")
        tmp = profile.runctx(
          command.__name__ + '(options=options,parser=parser)', command.__globals__, locals(), tfile
        )
        p = pstats.Stats(tfile).strip_dirs()
        p.sort_stats('time', 'cumulative')
        p = p.print_stats(options.profile)
        p.print_callers(options.profile)
        p.print_callees(options.profile)
        p = p.sort_stats('cumulative','calls')
        p.print_stats(options.profile)
        p.print_callers(options.profile)
        p.print_callees(options.profile)
        p = p.sort_stats('calls')
        p.print_stats(options.profile)
        p.print_callers(options.profile)
        p.print_callees(options.profile)
        retval = tmp
    else:
        #
        # Call the main Pyomo runner without profiling
        #
        TempfileManager.push()
        try:
            retval = command(options=options, parser=parser)
        except SystemExit:
            err = sys.exc_info()[1]
            if __debug__:
                if options.debug or options.catch:
                    sys.exit(0)
            print('Exiting %s: %s' % (name, str(err)))
            errorcode = err.code
        except Exception:
            err = sys.exc_info()[1]
            # If debugging is enabled, pass the exception up the chain
            # (to pyomo_excepthook)
            if __debug__:
                if options.debug or options.catch:
                    if not logfile is None:
                        pyutilib.misc.reset_redirect()
                    TempfileManager.pop()
                    raise

            if len(options.model_file) > 0:
                model = "model " + options.model_file
            else:
                model = "model"

            global filter_excepthook
            if filter_excepthook:
                action = "loading"
            else:
                action = "running"

            msg = "Unexpected exception while %s %s:\n" % (action, model)
            #
            # This handles the case where the error is propagated by a KeyError.
            # KeyError likes to pass raw strings that don't handle newlines
            # (they translate "\n" to "\\n"), as well as tacking on single
            # quotes at either end of the error message. This undoes all that.
            #
            errStr = str(err)
            if type(err) == KeyError and errStr != "None":
                errStr = str(err).replace(r"\n","\n")[1:-1]

            logging.getLogger('coopr.pyomo').error(msg+errStr)
            errorcode = 1

    if not logfile is None:
        pyutilib.misc.reset_redirect()

    if options.disable_gc:
        gc.enable()
    TempfileManager.pop(remove=not options.keepfiles)
    return Container(retval=retval, errorcode=errorcode)


def cleanup():
    for key in modelapi:
        for ep in ExtensionPoint(modelapi[key]):
            ep.deactivate()

        



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

#
# Problem Writer for CPLEX LP Format Files
#

import logging
import math
import weakref

from six import iterkeys, itervalues, iteritems, advance_iterator, StringIO
from six.moves import xrange, zip

from coopr.opt import ProblemFormat
from coopr.opt.base import AbstractProblemWriter
from coopr.pyomo.base import SymbolMap, BasicSymbolMap, TextLabeler, NumericLabeler
from coopr.pyomo.base import BooleanSet, Constraint, ConstraintList, expr, IntegerSet, Component
from coopr.pyomo.base import active_subcomponents_generator, active_subcomponents_data_generator
from coopr.pyomo.base import Var, value, label_from_name, NumericConstant, Suffix
from coopr.pyomo.base.sos import SOSConstraint
from coopr.pyomo.base.objective import Objective, minimize, maximize
from coopr.pyomo.expr import canonical_degree, LinearCanonicalRepn

from coopr.core.plugin import alias
from pyutilib.misc import tostr, PauseGC

logger = logging.getLogger('coopr.pyomo')

class ProblemWriter_cpxlp(AbstractProblemWriter):

    alias('cpxlp')
    alias('lp')

    def __init__(self):

        AbstractProblemWriter.__init__(self, ProblemFormat.cpxlp)

        # the LP writer is responsible for tracking which variables are
        # referenced in constraints, so that one doesn't end up with a
        # zillion "unreferenced variables" warning messages. stored at
        # the object level to avoid additional method arguments.
        # dictionary of id(_VarData)->_VarData.
        self._referenced_variable_ids = {}
        
        # Keven Hunter made a nice point about using %.16g in his attachment
        # to ticket #4319. I am adjusting this to %.17g as this mocks the
        # behavior of using %r (i.e., float('%r'%<number>) == <number>) with
        # the added benefit of outputting (+/-). The only case where this
        # fails to mock the behavior of %r is for large (long) integers (L),
        # which is a rare case to run into and is probably indicative of
        # other issues with the model.
        # *** NOTE ***: If you use 'r' or 's' here, it will break code that
        #               relies on using '%+' before the formatting character
        #               and you will need to go add extra logic to output
        #               the number's sign.
        self._precision_string = '.17g'

    def __call__(self, model, output_filename, solver_capability, io_options):

        # NOTE: io_options is a simple dictionary of keyword-value pairs
        #       specific to this writer. 
        symbolic_solver_labels = io_options.pop("symbolic_solver_labels", False)
        output_fixed_variable_bounds = io_options.pop("output_fixed_variable_bounds", False)
        labeler = io_options.pop("labeler", None)
        if io_options:
            logger.warn(
                "ProblemWriter_cpxlp passed unrecognized io_options:\n\t" +
                "\n\t".join("%s = %s" % (k,v) for k,v in iteritems(io_options)))

        if symbolic_solver_labels and (labeler is not None):
            raise ValueError("ProblemWriter_cpxlp: Using both the "
                             "'symbolic_solver_labels' and 'labeler' "
                             "I/O options is forbidden")

        if symbolic_solver_labels:
            labeler = TextLabeler()
        elif labeler is None:
            labeler = NumericLabeler('x')

        # when sorting, there are a non-trivial number of temporary objects
        # created. these all yield non-circular references, so disable GC -
        # the overhead is non-trivial, and because references are non-circular,
        # everything will be collected immediately anyway.
        suspend_gc = PauseGC()

        # clear the collection of referenced variables.
        self._referenced_variable_ids.clear()

        if output_filename is None:
            output_filename = model.name + ".lp"

        output_file=open(output_filename, "w")
        symbol_map = self._print_model_LP(model,
                                          output_file,
                                          solver_capability,
                                          labeler,
                                          output_fixed_variable_bounds)
        output_file.close()

        self._referenced_variable_ids.clear()

        return output_filename, symbol_map

    def _get_bound(self, exp):

        if exp.is_fixed():
            return exp()
        else:
            raise ValueError("ERROR: non-fixed bound: " + str(exp))

    def _print_expr_linear(self, x, output_file, object_symbol_dictionary, variable_symbol_dictionary, is_objective):

        """
        Return a expression as a string in LP format.

        Note that this function does not handle any differences in LP format
        interpretation by the solvers (e.g. CPlex vs GLPK).  That decision is
        left up to the caller.

        required arguments:
          x: A Pyomo linear encoding of an expression to write in LP format
        """

        # Currently, it appears that we only need to print the constant
        # offset term for objectives.
        print_offset = is_objective

        constant_term = x[0]
        linear_terms = x[1]

        name_to_coefficient_map = {}

        for coefficient, vardata in linear_terms:

            var_id = id(vardata)
            self._referenced_variable_ids[var_id] = vardata

            name = variable_symbol_dictionary[var_id]

            # due to potential disabling of expression simplification,
            # variables might appear more than once - condense coefficients.
            name_to_coefficient_map[name] = coefficient + name_to_coefficient_map.get(name,0.0)

        sorted_names = sorted(iterkeys(name_to_coefficient_map))

        string_template = '%+'+self._precision_string+' %s\n'
        for name in sorted_names:

            coefficient = name_to_coefficient_map[name]

            output_file.write(string_template % (coefficient, name))

        if print_offset and (constant_term != 0.0):

            output_file.write(string_template % (constant_term, 'ONE_VAR_CONSTANT'))

        return constant_term


    def _print_expr_canonical(self, x, output_file, object_symbol_dictionary, variable_symbol_dictionary, is_objective):

        """
        Return a expression as a string in LP format.

        Note that this function does not handle any differences in LP format
        interpretation by the solvers (e.g. CPlex vs GLPK).  That decision is
        left up to the caller.

        required arguments:
          x: A Pyomo canonical expression to write in LP format
        """

        # cache - this is referenced numerous times.
        if isinstance(x, LinearCanonicalRepn):
            var_hashes = None # not needed
        else:
            var_hashes = x[-1]

        #
        # Linear
        #
        linear_coef_string_template = '%+'+self._precision_string+' %s\n'
        if isinstance(x, LinearCanonicalRepn) and (x.linear is not None):

            # the 99% case is when the input instance is a linear canonical expression, so the exception should be rare.
            for i in xrange(0,len(x.linear)):
                var_coef = x.linear[i]
                vardata = x.variables[i]
                self._referenced_variable_ids[id(vardata)] = vardata

            sorted_names = [(variable_symbol_dictionary[id(x.variables[i])], x.linear[i]) for i in xrange(0,len(x.linear))]
            sorted_names.sort()

            for name, coef in sorted_names:
                output_file.write(linear_coef_string_template % (coef, name))
        elif 1 in x:

            for var_hash in x[1]:
                vardata = var_hashes[var_hash]
                self._referenced_variable_ids[id(vardata)] = vardata

            sorted_names = [(variable_symbol_dictionary[id(var_hashes[var_hash])], var_coefficient) for var_hash, var_coefficient in iteritems(x[1])]
            sorted_names.sort()

            for name, coef in sorted_names:
                output_file.write(linear_coef_string_template % (coef, name))

        #
        # Quadratic
        #
        quad_coef_string_template = '%+'+self._precision_string+' '
        if canonical_degree(x) is 2:

            # first, make sure there is something to output - it is possible for all
            # terms to have coefficients equal to 0.0, in which case you don't want
            # to get into the bracket notation at all.
            # NOTE: if the coefficient is really 0.0, it should be preprocessed out by
            #       the canonial expression generator!
            found_nonzero_term = False # until proven otherwise
            for var_hash, var_coefficient in iteritems(x[2]):
                for var in var_hash:
                    vardata = var_hashes[var]
                    self._referenced_variable_ids[id(vardata)] = vardata
                    
                if math.fabs(var_coefficient) != 0.0:
                    found_nonzero_term = True
                    break

            if found_nonzero_term:

                output_file.write("+ [\n")

                num_output = 0

                for var_hash in sorted(iterkeys(x[2])):

                    coefficient = x[2][var_hash]

                    if is_objective:
                        coefficient *= 2

                    # times 2 because LP format requires /2 for all the quadratic
                    # terms /of the objective only/.  Discovered the last bit thru
                    # trial and error.  Obnoxious.
                    # Ref: ILog CPlex 8.0 User's Manual, p197.

                    output_file.write(quad_coef_string_template % coefficient)
                    term_variables = []

                    for var in var_hash:
                        vardata = var_hashes[var]
                        name = variable_symbol_dictionary[id(vardata)]
                        term_variables.append(name)

                    if len(term_variables) == 2:
                        output_file.write("%s * %s" % (term_variables[0],term_variables[1]))
                    else:
                        output_file.write("%s ^ 2" % term_variables[0])
                    output_file.write("\n")

                output_file.write("]")

                if is_objective:
                    output_file.write(' / 2\n')
                    # divide by 2 because LP format requires /2 for all the quadratic
                    # terms.  Weird.  Ref: ILog CPlex 8.0 User's Manual, p197
                else:
                    output_file.write("\n")


        #
        # Constant offset
        #
        if isinstance(x, LinearCanonicalRepn):
            constant = x.constant
        else:
            if 0 in x:
                constant = x[0][None]
            else:
                constant = None

        if constant is not None:
            offset = constant
        else:
            offset=0.0

        # Currently, it appears that we only need to print the constant
        # offset term for objectives.
        obj_string_template = '%+'+self._precision_string+' %s\n'
        if is_objective and offset != 0.0:
            output_file.write(obj_string_template % (offset, 'ONE_VAR_CONSTANT'))

        #
        # Return constant offset
        #
        return offset

    def printSOS(self, symbol_map, labeler, variable_symbol_map, soscondata, output_file):
        """
        Prints the SOS constraint associated with the _SOSConstraintData object
        """

        sos_items = soscondata.get_items()
        level = soscondata.get_level()

        if len(sos_items) == 0:
            return

        output_file.write('%s: S%s::\n' % (symbol_map.getSymbol(soscondata,labeler), level))

        sos_template_string = "%s:%"+self._precision_string+"\n"
        for vardata, weight in sos_items:
            if vardata.fixed:
                raise RuntimeError("SOSConstraint '%s' includes a fixed variable '%s'. "
                                   "This is currently not supported. Deactive this constraint "
                                   "in order to proceed" % (soscondata.cname(True), vardata.cname(True)))
            self._referenced_variable_ids[id(vardata)] = vardata
            output_file.write(sos_template_string % (variable_symbol_map.getSymbol(vardata), weight) )

    #
    # a simple utility to pass through each variable and constraint
    # in the input model and populate the symbol map accordingly. once
    # we know all of the objects in the model, we can then "freeze"
    # the contents and use more efficient query mechanisms on simple
    # dictionaries - and avoid checking and function call overhead.
    #
    def _populate_symbol_map(self, model, symbol_map, labeler, variable_symbol_map):

        # NOTE: we use createSymbol instead of getSymbol because we know
        # whether or not the symbol exists, and don't want to the overhead
        # of error/duplicate checking.

        # cache frequently called functions
        create_symbol_func = SymbolMap.createSymbol
        create_symbols_func = SymbolMap.createSymbols
        alias_symbol_func = SymbolMap.alias
        variable_label_pairs = []

        for block in model.all_blocks(sort_by_keys=True):

            for objective_data in active_subcomponents_data_generator(block, Objective):
                create_symbol_func(symbol_map, objective_data, labeler)

            for constraint_data in active_subcomponents_data_generator(block, Constraint):

                constraint_data_symbol = create_symbol_func(symbol_map, constraint_data, labeler)
                if constraint_data._equality:
                    label = 'c_e_' + constraint_data_symbol + '_'
                    alias_symbol_func(symbol_map, constraint_data, label)
                else:
                    if constraint_data.lower is not None:
                        if constraint_data.upper is not None:
                            alias_symbol_func(symbol_map, constraint_data, 'r_l_' + constraint_data_symbol + '_')
                            alias_symbol_func(symbol_map, constraint_data, 'r_u_' + constraint_data_symbol + '_')
                        else:
                            label = 'c_l_' + constraint_data_symbol + '_'
                            alias_symbol_func(symbol_map, constraint_data, label)
                    elif constraint_data.upper is not None:
                        label = 'c_u_' + constraint_data_symbol + '_'
                        alias_symbol_func(symbol_map, constraint_data, label)
                        
            for con in active_subcomponents_data_generator(block,SOSConstraint):
                create_symbol_func(symbol_map, con, labeler)    

            variable_label_pairs.extend((var,create_symbol_func(symbol_map, var, labeler)) for var in active_subcomponents_data_generator(block, Var))

        variable_symbol_map.updateSymbols(variable_label_pairs)

    def _print_model_LP(self,
                        model,
                        output_file,
                        solver_capability,
                        labeler,
                        output_fixed_variable_bounds):

        symbol_map = SymbolMap(model)
        variable_symbol_map = BasicSymbolMap()

        # populate the symbol map in a single pass.
        self._populate_symbol_map(model, symbol_map, labeler, variable_symbol_map)

        # and extract the information we'll need for rapid labeling.
        object_symbol_dictionary = symbol_map.getByObjectDictionary()
        variable_symbol_dictionary = variable_symbol_map.byObject

        # cache - these are called all the time.
        print_expr_linear = self._print_expr_linear
        print_expr_canonical = self._print_expr_canonical

        # print the model name and the source, so we know roughly where
        # it came from.
        #
        # NOTE: this *must* use the "\* ... *\" comment format: the GLPK
        # LP parser does not correctly handle other formats (notably, "%").
        output_file.write(
            "\\* Source Pyomo model name=%s *\\\n\n" % (model.name,) )

        #
        # Objective
        #
        model_repn_suffix = model.subcomponent("canonical_repn")
        if (model_repn_suffix is None) or (not model_repn_suffix.type() is Suffix) or (not model_repn_suffix.active is True):
            raise ValueError("Unable to find an active Suffix with name 'canonical_repn' on block: %s" % (model.cname(True)))

        supports_quadratic_objective = solver_capability('quadratic_objective')

        numObj = 0
        onames = []
        for block in model.all_blocks(sort_by_keys=True):
            block_repn_suffix = block.subcomponent("canonical_repn")
            if (block_repn_suffix is None) or (not block_repn_suffix.type() is Suffix) or (not block_repn_suffix.active is True):
                block_repn_suffix = None
            #
            for objective in active_subcomponents_generator(block,Objective):
                for index in sorted(iterkeys(objective)):
                    objective_data = objective[index]
                    if not objective_data.active:
                        continue
                    #
                    numObj += 1
                    onames.append(objective_data.cname())
                    if numObj > 1:
                        msg = "More than one active objective defined for input model '%s'; " \
                              'Cannot write legal LP file\n'                           \
                              'Objectives: %s'
                        raise ValueError(
                            #msg % ( model.name,', '.join("'%s'" % x.cname(True) for x in _obj) ))
                            msg % ( model.name,' '.join(onames)))
                    #
                    symbol_map.alias(objective_data, '__default_objective__')
                    if objective_data.is_minimizing():
                        output_file.write("min \n")
                    else:
                        output_file.write("max \n")
                    #
                    obj_data_repn = block_repn_suffix.getValue(objective_data)
                    if obj_data_repn is None:
                        raise RuntimeError("No canonical representation identified for objective="+objective_data.cname())
                    degree = canonical_degree(obj_data_repn)
                    #
                    if degree == 0:
                        print("Warning: Constant objective detected, replacing " +
                               "with a placeholder to prevent solver failure.")
                        output_file.write(object_symbol_dictionary[id(objective_data)] + ": +0.0 ONE_VAR_CONSTANT\n")
                        # Skip the remaining logic of the section
                    else:
                        if degree == 2:
                            if not supports_quadratic_objective:
                                raise RuntimeError(
                                    'Selected solver is unable to handle objective functions with quadratic terms. ' \
                                    'Objective at issue: %s.' % objective_data.cname())
                        elif degree != 1:
                            msg  = "Cannot write legal LP file.  Objective '%s' "  \
                                   'has nonlinear terms that are not quadratic.' % objective_data.cname(True)
                            raise RuntimeError(msg)

                        output_file.write(object_symbol_dictionary[id(objective_data)]+':\n')

                        offset = print_expr_canonical( obj_data_repn,
                                                       output_file,
                                                       object_symbol_dictionary,
                                                       variable_symbol_dictionary,
                                                       True )

        if numObj == 0:
            msg = "ERROR: No objectives defined for input model '%s'; "    \
                  ' cannot write legal LP file'
            raise ValueError(msg % str( model.name ))


        # Constraints
        #
        # If there are no non-trivial constraints, you'll end up with an empty
        # constraint block. CPLEX is OK with this, but GLPK isn't. And
        # eliminating the constraint block (i.e., the "s.t." line) causes GLPK
        # to whine elsewhere. output a warning if the constraint block is empty,
        # so users can quickly determine the cause of the solve failure.

        output_file.write("\n")
        output_file.write("s.t.\n")
        output_file.write("\n")

        have_nontrivial = False

        supports_quadratic_constraint = solver_capability('quadratic_constraint')

        # FIXME: This is a hack to get nested blocks working...
        eq_string_template = "= %"+self._precision_string+'\n'
        geq_string_template = ">= %"+self._precision_string+'\n\n'
        leq_string_template = "<= %"+self._precision_string+'\n\n'
        for block in model.all_blocks(sort_by_keys=True):

            block_repn_suffix = block.subcomponent("canonical_repn")
            if (block_repn_suffix is None) or (not block_repn_suffix.type() is Suffix) or (not block_repn_suffix.active is True):
                block_repn_suffix = None
            block_lin_body_suffix = block.subcomponent("lin_body")
            if (block_lin_body_suffix is None) or (not block_lin_body_suffix.type() is Suffix) or (not block_lin_body_suffix.active is True):
                block_lin_body_suffix = None

            for constraint in active_subcomponents_generator(block,Constraint):

                if constraint.trivial:
                    continue

                have_nontrivial=True

                for index in sorted(iterkeys(constraint)):

                    constraint_data = constraint[index]
                    if not constraint_data.active:
                        continue

                    # if expression trees have been linearized, then the canonical
                    # representation attribute on the constraint data object will
                    # be equal to None.
                    constraint_data_repn = None
                    if block_repn_suffix is not None:
                        constraint_data_repn = block_repn_suffix.getValue(constraint_data)
                    if constraint_data_repn is not None:

                        degree = canonical_degree(constraint_data_repn)

                        # There are conditions, e.g., when fixing variables, under which
                        # a constraint block might be empty.  Ignore these, for both
                        # practical reasons and the fact that the CPLEX LP format
                        # requires a variable in the constraint body.  It is also
                        # possible that the body of the constraint consists of only a
                        # constant, in which case the "variable" of
                        if degree == 0:
                            # this happens *all* the time in many applications,
                            # including PH - so suppress the warning.
                            #
                            #msg = 'WARNING: ignoring constraint %s[%s] which is ' \
                            #      'constant'
                            #print msg % (str(C),str(index))
                            continue

                        if degree == 2:
                            if not supports_quadratic_constraint:
                                msg  = 'Solver unable to handle quadratic expressions.'\
                                       "  Constraint at issue: '%s%%s'"
                                msg %= constraint.name
                                if index is None:
                                    msg %= ''
                                else:
                                    msg %= '%s' % ( str(index)
                                        .replace('(', '[')
                                        .replace(')', ']')
                                    )

                                raise ValueError(msg)

                        elif degree != 1:
                            msg = "Cannot write legal LP file.  Constraint '%s%s' "   \
                                  'has a body with nonlinear terms.'
                            if index is None:
                                msg %= ( constraint.name, '')
                            else:
                                msg %= ( constraint.name, '[%s]' % index )
                            raise ValueError(msg)

                    con_symbol = object_symbol_dictionary[id(constraint_data)]
                    if constraint_data._equality:
                        label = 'c_e_' + con_symbol + '_'
                        output_file.write(label+':\n')
                        lin_body = None
                        if block_lin_body_suffix is not None:
                            lin_body = block_lin_body_suffix.getValue(constraint_data)
                        if lin_body is not None:
                            offset = print_expr_linear(lin_body, output_file, object_symbol_dictionary, variable_symbol_dictionary, False)
                        else:
                            offset = print_expr_canonical(constraint_data_repn, output_file, object_symbol_dictionary, variable_symbol_dictionary, False)
                        bound = constraint_data.lower
                        bound = self._get_bound(bound) - offset
                        output_file.write(eq_string_template%bound)
                        output_file.write("\n")
                    else:
                        lin_body = None
                        if block_lin_body_suffix is not None:
                            lin_body = block_lin_body_suffix.getValue(constraint_data)
                        if constraint_data.lower is not None:
                            if constraint_data.upper is not None:
                                label = 'r_l_' + con_symbol + '_'
                            else:
                                label = 'c_l_' + con_symbol + '_'
                            output_file.write(label+':\n')
                            if lin_body is not None:
                                offset = print_expr_linear(lin_body, output_file, object_symbol_dictionary, variable_symbol_dictionary, False)
                            else:
                                offset = print_expr_canonical(constraint_data_repn, output_file, object_symbol_dictionary, variable_symbol_dictionary, False)
                            bound = constraint_data.lower
                            bound = self._get_bound(bound) - offset
                            output_file.write(geq_string_template%bound)
                        if constraint_data.upper is not None:
                            if constraint_data.lower is not None:
                                label = 'r_u_' + con_symbol + '_'
                            else:
                                label = 'c_u_' + con_symbol + '_'
                            output_file.write(label+':\n')
                            if lin_body is not None:
                                offset = print_expr_linear(lin_body, output_file, object_symbol_dictionary, variable_symbol_dictionary, False)
                            else:
                                offset = print_expr_canonical(constraint_data_repn, output_file, object_symbol_dictionary, variable_symbol_dictionary, False)
                            bound = constraint_data.upper
                            bound = self._get_bound(bound) - offset
                            output_file.write(leq_string_template%bound)

        if not have_nontrivial:
            print('WARNING: Empty constraint block written in LP format '  \
                  '- solver may error')

        # the CPLEX LP format doesn't allow constants in the objective (or
        # constraint body), which is a bit silly.  To avoid painful
        # book-keeping, we introduce the following "variable", constrained
        # to the value 1.  This is used when quadratic terms are present.
        # worst-case, if not used, is that CPLEX easily pre-processes it out.
        prefix = ""
        output_file.write('%sc_e_ONE_VAR_CONSTANT: \n' % prefix)
        output_file.write('%sONE_VAR_CONSTANT = 1.0\n' % prefix)
        output_file.write("\n")

        # SOS constraints
        #
        # For now, we write out SOS1 and SOS2 constraints in the cplex format
        #
        # All Component objects are stored in model._component, which is a
        # dictionary of {class: {objName: object}}.
        #
        # Consider the variable X,
        #
        #   model.X = Var(...)
        #
        # We print X to CPLEX format as X(i,j,k,...) where i, j, k, ... are the
        # indices of X.
        #
        SOSlines = StringIO()
        sos1 = solver_capability("sos1")
        sos2 = solver_capability("sos2")
        writtenSOS = False
        for block in model.all_blocks(sort_by_keys=True):
            for soscondata in active_subcomponents_data_generator(block,SOSConstraint):
                level = soscondata.get_level()
                if (level == 1 and not sos1) or (level == 2 and not sos2) or (level > 2):
                    raise ValueError("Solver does not support SOS level %s constraints"
                                     % (level,) )
                if writtenSOS == False:
                    SOSlines.write("SOS\n")
                    writtenSOS = True
                # This updates the referenced_variable_ids, just in case there is a variable
                # that only appears in an SOSConstraint, in which case this needs to be known
                # before we write the "bounds" section (Cplex does not handle this correctly, Gurobi does)
                self.printSOS(symbol_map, labeler, variable_symbol_map, soscondata, SOSlines)

        #
        # Bounds
        #

        output_file.write("bounds \n")

        # Scan all variables even if we're only writing a subset of them.
        # required because we don't store maps by variable type currently.

        # FIXME: This is a hack to get nested blocks working...
        lb_string_template = "%"+self._precision_string+" <= "
        ub_string_template = " <= %"+self._precision_string+"\n"
        # Track the number of integer and binary variables, so you can
        # output their status later.
        integer_vars = []
        binary_vars = []
        for block in model.all_blocks(sort_by_keys=True):
            for vardata in active_subcomponents_data_generator(block, Var, sort_by_keys=True):

                # TODO: We could just loop over the set of items in self._referenced_variable_ids,
                #       except this is a dictionary that is hashed by id(vardata) which would make
                #       the bounds section nondeterministic (bad for unit testing)
                if id(vardata) not in self._referenced_variable_ids:
                    continue

                if vardata.fixed:
                    if not output_fixed_variable_bounds:
                        raise ValueError("Encountered a fixed variable (%s) inside an active objective "
                                         "or constraint expression on model %s, which is usually indicative of "
                                         "a preprocessing error. Use the IO-option 'output_fixed_variable_bounds=True' "
                                         "to suppress this error and fix the variable by overwriting its bounds in "
                                         "the LP file."
                                         % (vardata.cname(True),model.cname(True),))
                    if vardata.value is None:
                        raise ValueError("Variable cannot be fixed to a value of None.")
                    vardata_lb = vardata.value
                    vardata_ub = vardata.value
                else:
                    vardata_lb = vardata.lb
                    vardata_ub = vardata.ub

                name_to_output = variable_symbol_dictionary[id(vardata)]

                # track the number of integer and binary variables, so we know whether
                # to output the general / binary sections below.
                if vardata.is_integer():
                    integer_vars.append(name_to_output)
                elif vardata.is_binary():
                    binary_vars.append(name_to_output)
                elif not vardata.is_continuous():
                    raise TypeError("Invalid domain type for variable with name '%s'. "
                                    "Variable is not continuous, integer, or binary.")

                # in the CPLEX LP file format, the default variable
                # bounds are 0 and +inf.  These bounds are in
                # conflict with Pyomo, which assumes -inf and +inf
                # (which we would argue is more rational).
                output_file.write("   ")
                if vardata_lb is not None:
                    output_file.write(lb_string_template % value(vardata_lb))
                else:
                    output_file.write(" -inf <= ")
                if name_to_output == "e":
                    msg = 'Attempting to write variable with name' \
                        "'e' in a CPLEX LP formatted file - "    \
                        'will cause a parse failure due to '     \
                        'confusion with numeric values '         \
                        'expressed in scientific notation'
                    raise ValueError(msg)
                output_file.write(name_to_output)
                if vardata_ub is not None:
                    output_file.write(ub_string_template % value(vardata_ub))
                else:
                    output_file.write(" <= +inf\n")

        if len(integer_vars) > 0:

            output_file.write("general\n")
            for var_name in integer_vars:
                output_file.write('  %s\n' % var_name)

        if len(binary_vars) > 0:

            output_file.write("binary\n")
            for var_name in binary_vars:
                output_file.write('  %s\n' % var_name)


        # Write the SOS section
        output_file.write(SOSlines.getvalue())

        #
        # wrap-up
        #
        output_file.write("end \n")

        # Clean up the symbol map to only contain variables referenced in the active constraints
        # **Note**: warm start method may rely on this for choosing the set of potential warm start variables 
        vars_to_delete = set(variable_symbol_map.byObject.keys())-set(self._referenced_variable_ids.keys())
        sm_byObject = symbol_map.byObject
        sm_bySymbol = symbol_map.bySymbol
        var_sm_byObject = variable_symbol_map.byObject
        for varid in vars_to_delete:
            symbol = var_sm_byObject[varid]
            del sm_byObject[varid]
            del sm_bySymbol[symbol]
        del variable_symbol_map

        return symbol_map



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the FAST README.txt file.
#  _________________________________________________________________________
from __future__ import division

__all__ = [ 'AmplRepn', 'generate_ampl_repn']

import weakref

from coopr.pyomo.base import expr as Expr
from coopr.pyomo.base import _ExpressionData, Expression
from coopr.pyomo.base.var import _VarData, value
from coopr.pyomo.base import external
from coopr.pyomo.expr.canonical_repn import collect_linear_canonical_repn

import six
from six import itervalues, iteritems, StringIO
from six.moves import xrange, zip

import logging

logger = logging.getLogger('coopr.pyomo')

using_py3 = six.PY3

class AmplRepn(object):

    __slots__ = ('_constant',
                 '_linear_terms_coef',
                 '_linear_vars',
                 '_nonlinear_expr',
                 '_nonlinear_vars')

    def __getstate__(self):
        """
        This method is required because this class uses slots.
        """ 
        return  (self._constant,
                 self._linear_terms_coef,
                 self._linear_vars,
                 self._nonlinear_expr,
                 self._nonlinear_vars)

    def __setstate__(self, state):
        """
        This method is required because this class uses slots.
        """
        self._constant, \
        self._linear_terms_coef, \
        self._linear_vars, \
        self._nonlinear_expr, \
        self._nonlinear_vars = state
        
    def __init__(self):
        self._constant = 0
        self._linear_vars = {}
        self._linear_terms_coef = {}
        self._nonlinear_expr = None
        self._nonlinear_vars = {}

    # 
    # Although it is convenient to have the dictionaries hashed by some
    # variable id when generating the ampl representation, we can 
    # compress these to lists after generation is complete
    #
    def compress(self):
        if  self._linear_vars.__class__ is dict:
            linear_keys = self._linear_vars.keys()
            self._linear_vars = tuple(self._linear_vars[key] for key in linear_keys)
            self._linear_terms_coef = tuple(self._linear_terms_coef[key] for key in linear_keys)
        if self._nonlinear_vars.__class__ is dict:
            self._nonlinear_vars = tuple(itervalues(self._nonlinear_vars))

    # GAH: I'm not sure why this funciton is here and I think it is misleading. Calling clone on 
    #      a model will deepcopy the AmplRepn (e.g. new expressions and variables). Calling clone
    #      directly on the AmplRepn (as defined below) does nothing close to that.
    def Xclone(self):
        clone = AmplRepn()
        clone._constant = self._constant
        clone._linear_vars = self._linear_vars
        clone._linear_terms_coef = self._linear_terms_coef
        clone._nonlinear_expr = self._nonlinear_expr
        clone._nonlinear_vars = self._nonlinear_vars
        return clone

    # for debugging
    def __str__(self):
        output = StringIO()
        output.write("\n")
        output.write("constant:       "+str(self._constant)+"\n")
        output.write("linear vars:    "+str(self._linear_vars)+"\n")
        output.write("linear coef:    "+str(self._linear_terms_coef)+"\n")
        output.write("nonlinear expr:\n")
        try:
            self._nonlinear_expr.pprint(ostream=output)
        except AttributeError:
            output.write(str(self._nonlinear_expr)+"\n")
        output.write("nonlinear vars: "+str(self._nonlinear_vars)+"\n")
        output.write("\n")
        ret_str = output.getvalue()
        output.close()
        return ret_str

    def __eq__(self, other):
        # Can only be equal to other AmplRepn instances
        if not isinstance(other, AmplRepn):
            return False
        
        # Immediately check constants
        if self._constant != other._constant:
            return False

        # Optimization: check linear term lengths before iterating
        if len(self._linear_vars) != len(other._linear_vars):
            return False

        if self._linear_vars.__class__ is dict:
            self_linear_vars = list(itervalues(self._linear_vars))
            self_linear_terms = self._linear_terms_coef
        else:
            self_linear_vars = self._linear_vars
            self_linear_terms = dict((id(var),coef) for var,coef in zip(self._linear_vars,self._linear_terms_coef))

        if other._linear_vars.__class__ is dict:
            other_linear_vars = list(itervalues(other._linear_vars))
            other_linear_terms = other._linear_terms_coef
        else:
            other_linear_vars = other._linear_vars
            other_linear_terms = dict((id(var),coef) for var,coef in zip(other._linear_vars,other._linear_terms_coef))

        # Establish a mapping between self's linear terms and other's
        found_match = []
        for var in self_linear_vars:
            match = False
            for other_var in other_linear_vars:
                if var is other_var:
                    match = True
                    break
            found_match.append(match)
                    
        # We have to have found all our own vars
        if not all(found_match):
            return False

        # Check that linear term coefficients are equal
        for var_ID in self_linear_terms:
            if self_linear_terms[var_ID] != other_linear_terms[var_ID]:
                return False

        if self._nonlinear_expr is not None:
            if other._nonlinear_expr is None:
                return False

            if type(self._nonlinear_expr) is list:
                if type(other._nonlinear_expr) is not list:
                    return False

                # Check that nonlinear expressions are the same length
                if len(self._nonlinear_expr) != len(other._nonlinear_expr):
                    return False

                # Check that nonlinear expression have the same terms
                for i in xrange(len(self._nonlinear_expr)):
                    if self._nonlinear_expr[i][0] != other._nonlinear_expr[i][0]:
                        return False
                    if not self._nonlinear_expr[i][1] is other._nonlinear_expr[i][1]:
                        return False
            else:
                if not self._nonlinear_expr is other._nonlinear_expr:
                    return False
        else:
            if other._nonlinear_expr is not None:
                return False

        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def is_fixed(self):
        if len(self._linear_vars) == 0 and len(self._nonlinear_vars) == 0:
            return True
        return False

    def is_linear(self):
        if self._nonlinear_expr is None:
            return True
        return False

    def is_nonlinear(self):
        if self._nonlinear_expr is None:
            return False
        return True


def _generate_ampl_repn(exp):
    ampl_repn = AmplRepn()

    #
    # Expression
    #
    if exp.is_expression():
        exp_type = type(exp)

        #
        # Sum
        #
        if exp_type is Expr._SumExpression:
            ampl_repn._constant = exp._const #float(exp._const)
            ampl_repn._nonlinear_expr = None
            for i in xrange(len(exp._args)):
                exp_coef = exp._coef[i]
                child_repn = _generate_ampl_repn(exp._args[i])
                # adjust the constant
                ampl_repn._constant += exp_coef*child_repn._constant

                # adjust the linear terms
                for var_ID in child_repn._linear_vars:
                    if var_ID in ampl_repn._linear_terms_coef:
                        ampl_repn._linear_terms_coef[var_ID] += exp_coef*child_repn._linear_terms_coef[var_ID]
                    else:
                        ampl_repn._linear_terms_coef[var_ID] = exp_coef*child_repn._linear_terms_coef[var_ID]
                # adjust the linear vars
                ampl_repn._linear_vars.update(child_repn._linear_vars)

                # adjust the nonlinear terms
                if not child_repn._nonlinear_expr is None:
                    if ampl_repn._nonlinear_expr is None:
                        ampl_repn._nonlinear_expr = [(exp_coef, child_repn._nonlinear_expr)]
                    else:
                        ampl_repn._nonlinear_expr.append((exp_coef, child_repn._nonlinear_expr))
                # adjust the nonlinear vars
                ampl_repn._nonlinear_vars.update(child_repn._nonlinear_vars)
            return ampl_repn

        #
        # Product
        #
        elif exp_type is Expr._ProductExpression:
            #
            # Iterate through the denominator.  If they aren't all constants, then
            # simply return this expresion.
            #
            denom=1.0
            for e in exp._denominator:
                if e.is_fixed():
                    denom *= value(e)
                else:
                    ampl_repn._nonlinear_expr = exp
                    break
                if denom == 0.0:
                    print("Divide-by-zero error - offending sub-expression:")
                    e.pprint()
                    raise ZeroDivisionError

            if not ampl_repn._nonlinear_expr is None:
                # we have a nonlinear expression ... build up all the vars
                for e in exp._denominator:
                    arg_repn = _generate_ampl_repn(e)
                    ampl_repn._nonlinear_vars.update(arg_repn._linear_vars)
                    ampl_repn._nonlinear_vars.update(arg_repn._nonlinear_vars)

                for e in exp._numerator:
                    arg_repn = _generate_ampl_repn(e)
                    ampl_repn._nonlinear_vars.update(arg_repn._linear_vars)
                    ampl_repn._nonlinear_vars.update(arg_repn._nonlinear_vars)
                return ampl_repn

            #
            # OK, the denominator is a constant.
            #
            # build up the ampl_repns for the numerator
            
            n_linear_args = 0
            n_nonlinear_args = 0
            arg_repns = list()
            for i in xrange(len(exp._numerator)):
                e = exp._numerator[i]
                e_repn = _generate_ampl_repn(e)
                arg_repns.append(e_repn)
                # check if the expression is not nonlinear else it is nonlinear
                if e_repn._nonlinear_expr is not None:
                    n_nonlinear_args += 1
                # Check whether the expression is constant or else it is linear
                elif len(e_repn._linear_vars) > 0:
                    n_linear_args += 1
                # At this point we do not have a nonlinear expression and there are
                # no linear terms. If the expression constant is zero, then we
                # have a zero term in the product expression, so the entire product 
                # expression becomes trivial. 
                elif e_repn._constant == 0.0:
                    ampl_repn = e_repn
                    return ampl_repn

            is_nonlinear = False
            if n_linear_args > 1 or n_nonlinear_args > 0:
                is_nonlinear = True


            if is_nonlinear is True:
                # do like AMPL and simply return the expression
                # without extracting the potentially linear part
                ampl_repn = AmplRepn()
                ampl_repn._nonlinear_expr = exp
                for repn in arg_repns:
                    ampl_repn._nonlinear_vars.update(repn._linear_vars)
                    ampl_repn._nonlinear_vars.update(repn._nonlinear_vars)
                return ampl_repn

            else: # is linear or constant
                ampl_repn = current_repn = arg_repns[0]
                for i in xrange(1,len(arg_repns)):
                    e_repn = arg_repns[i]
                    ampl_repn = AmplRepn()

                    # const_c * const_e
                    ampl_repn._constant = current_repn._constant * e_repn._constant

                    # const_e * L_c
                    if e_repn._constant != 0.0:
                        for (var_ID, var) in iteritems(current_repn._linear_vars):
                            ampl_repn._linear_terms_coef[var_ID] = current_repn._linear_terms_coef[var_ID] * e_repn._constant
                        ampl_repn._linear_vars.update(current_repn._linear_vars)

                    # const_c * L_e
                    if current_repn._constant != 0.0:
                        for (e_var_ID,e_var) in iteritems(e_repn._linear_vars):
                            if e_var_ID in ampl_repn._linear_vars:
                                ampl_repn._linear_terms_coef[e_var_ID] += current_repn._constant * e_repn._linear_terms_coef[e_var_ID]
                            else:
                                ampl_repn._linear_terms_coef[e_var_ID] = current_repn._constant * e_repn._linear_terms_coef[e_var_ID]
                        ampl_repn._linear_vars.update(e_repn._linear_vars)
                    current_repn = ampl_repn

            # now deal with the product expression's coefficient that needs
            # to be applied to all parts of the ampl_repn
            ampl_repn._constant *= exp._coef/denom
            for var_ID in ampl_repn._linear_terms_coef:
                ampl_repn._linear_terms_coef[var_ID] *= exp._coef/denom

            return ampl_repn

        #
        # Power Expressions
        #
        elif exp_type is Expr._PowExpression:
            assert(len(exp._args) == 2)
            base_repn = _generate_ampl_repn(exp._args[0])
            base_repn_fixed = base_repn.is_fixed()
            exponent_repn = _generate_ampl_repn(exp._args[1])
            exponent_repn_fixed = exponent_repn.is_fixed()

            if base_repn_fixed and exponent_repn_fixed:
                ampl_repn._constant = base_repn._constant**exponent_repn._constant
            elif exponent_repn_fixed and exponent_repn._constant == 1.0:
                ampl_repn = base_repn
            elif exponent_repn_fixed and exponent_repn._constant == 0.0:
                ampl_repn._constant = 1.0
            else:
                # instead, let's just return the expression we are given and only
                # use the ampl_repn for the vars
                ampl_repn._nonlinear_expr = exp
                ampl_repn._nonlinear_vars = base_repn._nonlinear_vars
                ampl_repn._nonlinear_vars.update(exponent_repn._nonlinear_vars)
                ampl_repn._nonlinear_vars.update(base_repn._linear_vars)
                ampl_repn._nonlinear_vars.update(exponent_repn._linear_vars)
            return ampl_repn
        #
        # External Functions
        #
        elif exp_type is external._ExternalFunctionExpression:
            # In theory, the argument are fixed, we can simply evaluate this expression
            if exp.is_fixed():
                ampl_repn._constant = value(exp)
                return ampl_repn

            # this is inefficient since it is using much more than what we need
            ampl_repn._nonlinear_expr = exp
            for arg in exp._args:
                if isinstance(arg, basestring):
                    continue
                child_repn = _generate_ampl_repn(arg)
                ampl_repn._nonlinear_vars.update(child_repn._nonlinear_vars)
                ampl_repn._nonlinear_vars.update(child_repn._linear_vars)
            return ampl_repn
        #
        # Intrinsic Functions
        #
        elif isinstance(exp,Expr._IntrinsicFunctionExpression):
            # Checking isinstance above accounts for the fact that _AbsExpression
            # is a subclass of _IntrinsicFunctionExpression
            assert(len(exp._args) == 1)
            
            # the argument is fixed, we can simply evaluate this expression
            if exp._args[0].is_fixed():
                ampl_repn._constant = value(exp)
                return ampl_repn

            # this is inefficient since it is using much more than what we need
            child_repn = _generate_ampl_repn(exp._args[0])

            ampl_repn._nonlinear_expr = exp
            ampl_repn._nonlinear_vars = child_repn._nonlinear_vars
            ampl_repn._nonlinear_vars.update(child_repn._linear_vars)
            return ampl_repn
        #
        # AMPL-style If-Then-Else expression
        #
        elif exp_type is Expr.Expr_if:
            if exp._if.is_fixed():
                if exp._if():
                    ampl_repn = _generate_ampl_repn(exp._then)
                else:
                    ampl_repn = _generate_ampl_repn(exp._else)
            else:
                if_repn = _generate_ampl_repn(exp._if)
                then_repn = _generate_ampl_repn(exp._then)
                else_repn = _generate_ampl_repn(exp._else)
                ampl_repn._nonlinear_expr = exp
                ampl_repn._nonlinear_vars = if_repn._nonlinear_vars
                ampl_repn._nonlinear_vars.update(then_repn._nonlinear_vars)
                ampl_repn._nonlinear_vars.update(else_repn._nonlinear_vars)
                ampl_repn._nonlinear_vars.update(if_repn._linear_vars)
                ampl_repn._nonlinear_vars.update(then_repn._linear_vars)
                ampl_repn._nonlinear_vars.update(else_repn._linear_vars)
            return ampl_repn
        elif (exp_type is Expr._InequalityExpression) or (exp_type is Expr._EqualityExpression):
            for arg in exp._args:
                arg_repn = _generate_ampl_repn(arg)
                ampl_repn._nonlinear_vars.update(arg_repn._nonlinear_vars)
                ampl_repn._nonlinear_vars.update(arg_repn._linear_vars)
            ampl_repn._nonlinear_expr = exp
            return ampl_repn
        elif exp.is_fixed():
            ampl_repn._constant = value(exp)
            return ampl_repn
        #
        # Expression (the component)
        #
        elif type(exp) is _ExpressionData or exp.type() is Expression:
            ampl_repn = _generate_ampl_repn(exp.value)
            return ampl_repn
        #
        # ERROR
        #
        else:
            raise ValueError("Unsupported expression type: "+str(exp))
    #
    # Constant
    #
    elif exp.is_fixed():
        ### GAH: Why were we even checking this
        #if not exp.value.__class__ in native_numeric_types:
        #    ampl_repn = _generate_ampl_repn(exp.value) 
        #    return ampl_repn
        ampl_repn._constant = value(exp)
        return ampl_repn
    #
    # Variable
    #
    elif isinstance(exp, _VarData):
        if exp.fixed:
            ampl_repn._constant = exp.value
            return ampl_repn
        var_ID = id(exp)
        ampl_repn._linear_terms_coef[var_ID] = 1.0
        ampl_repn._linear_vars[var_ID] = exp
        return ampl_repn

    #
    # ERROR
    #
    else:
        raise ValueError("Unexpected expression type: "+str(exp))

def generate_ampl_repn(exp):
    degree = exp.polynomial_degree()
    if (degree is None) or (degree > 1):
        repn = _generate_ampl_repn(exp)
        repn.compress()
    elif degree == 0:
        repn = AmplRepn()
        repn._constant = value(exp)
        # compress
        repn._linear_vars = tuple()
        repn._linear_terms_coef = tuple()
        repn._nonlinear_vars = tuple()
    else: # degree == 1
        repn = AmplRepn()
        # compress
        repn._linear_vars = tuple()
        repn._linear_terms_coef = tuple()
        repn._nonlinear_vars = tuple()
        coef, varmap = collect_linear_canonical_repn(exp, {})
        if None in coef:
            val = coef.pop(None)
            if val:
                repn._constant = val
        # the six module is inefficient in terms of wrapping iterkeys and itervalues,
        # in the context of Python 2.7. use the native dictionary methods where possible.
        if using_py3 is False:
            repn._linear_terms_coef = tuple(val for val in coef.itervalues() if val)
            repn._linear_vars = tuple((varmap[var_hash] for var_hash,val in coef.iteritems() if val))
        else:
            repn._linear_terms_coef = tuple(val for val in coef.values() if val)
            repn._linear_vars = tuple((varmap[var_hash] for var_hash,val in coef.items() if val))
    return repn



#
# Author:  Gabe Hackebeil
# Purpose: For regression testing to ensure that the Pyomo
#          NL writer properly reclassifies nonlinear expressions
#          as linear or trivial when fixing variables or params
#          cause such a situation.
#
#          This test model relies on the asl_test executable. It
#          will not solve if sent to a real optimizer.
#

from coopr.pyomo import *

model = ConcreteModel()

model.x = Var()
model.y = Var(initialize=0.0)
model.z = Var()
model.p = Param(initialize=0.0,mutable=True)
model.q = Param(initialize=0.0,mutable=False)

model.y.fixed = True

model.obj = Objective( expr=model.x )

model.con1 = Constraint(expr= model.x*model.y*model.z + model.x == 1.0)
model.con2 = Constraint(expr= model.x*model.p*model.z + model.x == 1.0)
model.con3 = Constraint(expr= model.x*model.q*model.z + model.x == 1.0)
# Pyomo differs from AMPL in these cases that involve immutable params (q).
# These never actually become constraints in Pyomo, and for good reason.
model.con4 = Constraint(expr= model.x*model.y*model.z == 1.0)
model.con5 = Constraint(expr= model.x*model.p*model.z == 1.0)
model.con6 = Constraint(rule= simple_constraint_rule(model.x*model.q*model.z == 0.0) )



### Pyomo sub-problem for AMPL Bender's example, available
### from http://www.ampl.com/NEW/LOOP2/stoch2.mod

from coopr.pyomo import *

model = AbstractModel()
model.name = "SubProblem"

# Declare suffixes
model.rc = Suffix(direction=Suffix.IMPORT)
model.lrc = Suffix(direction=Suffix.IMPORT)
model.urc = Suffix(direction=Suffix.IMPORT)
model.dual = Suffix(direction=Suffix.IMPORT)

# products
model.PROD = Set()

# number of weeks
model.T = Param(within=PositiveIntegers)

# derived set containing all valid week indices and subsets of interest.
def weeks_rule(model):
    return set(sequence(model.T()))
model.WEEKS = Set(initialize=weeks_rule, within=PositiveIntegers)

def two_plus_weeks_rule(model):
    return set(sequence(2, model.T()))
model.TWOPLUSWEEKS = Set(initialize=two_plus_weeks_rule, within=PositiveIntegers)

def three_plus_weeks_rule(model):
    return set(sequence(3, model.T()))
model.THREEPLUSWEEKS = Set(initialize=three_plus_weeks_rule, within=PositiveIntegers)

# tons per hour produced
model.rate = Param(model.PROD, within=PositiveReals)

# hours available in week
model.avail = Param(model.WEEKS, within=NonNegativeReals)

# limit on tons sold in week
model.market = Param(model.PROD, model.WEEKS, within=NonNegativeReals)

# cost per ton produced
model.prodcost = Param(model.PROD, within=NonNegativeReals)

# carrying cost/ton of inventory
model.invcost = Param(model.PROD, within=NonNegativeReals)

# projected revenue/ton
model.revenue = Param(model.PROD, model.WEEKS, within=NonNegativeReals)

# scenario probability
def unit_interval_validate(model, value):
    return (value >= 0.0) and (value <= 1.0)
model.prob = Param(validate=unit_interval_validate)

# inventory at end of first period.
model.inv1 = Param(model.PROD, within=NonNegativeReals, mutable=True)

# tons produced
model.Make = Var(model.PROD, model.TWOPLUSWEEKS, domain=NonNegativeReals)

# tons inventoried
model.Inv = Var(model.PROD, model.TWOPLUSWEEKS, domain=NonNegativeReals)

# tons sold
def sell_bounds(model, p, t):
    return (0, model.market[p, t])
model.Sell = Var(model.PROD, model.TWOPLUSWEEKS, within=NonNegativeReals, bounds=sell_bounds)

def time_rule(model, t):
    return sum([(1.0 / model.rate[p]) * model.Make[p, t] for p in model.PROD]) - model.avail[t] <= 0.0
model.Time = Constraint(model.TWOPLUSWEEKS, rule=time_rule)

def balance2_rule(model, p):
    return (model.Make[p, 2] + model.inv1[p]) - (model.Sell[p, 2] + model.Inv[p, 2]) == 0.0
model.Balance2 = Constraint(model.PROD, rule=balance2_rule)

def balance_rule(model, p, t):
    return (model.Make[p, t] + model.Inv[p, t-1]) - (model.Sell[p, t] + model.Inv[p, t]) == 0.0
model.Balance = Constraint(model.PROD, model.THREEPLUSWEEKS, rule=balance_rule)

# the manual distribution of model.prob is ugly, but at the moment necessary; Pyomo
# expression simplification will be significantly improved in the near-term future.
def exp_stage2_profit_rule(model):
    return sum([model.prob * model.revenue[p, t] * model.Sell[p, t] - \
                model.prob * model.prodcost[p] * model.Make[p, t] - \
                model.prob * model.invcost[p] * model.Inv[p, t] \
                for p in model.PROD for t in model.TWOPLUSWEEKS])
model.Exp_Stage2_Profit = Objective(rule=exp_stage2_profit_rule, sense=maximize)



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

import os.path
import re

from pyutilib.misc import Options
from coopr.core.plugin import alias, Plugin, implements

from coopr.pyomo.base.plugin import IDataManager
from coopr.pyomo.data.process_data import _process_include


class AmplDataCommands(Plugin):

    alias("dat", "Import data with AMPL data commands.")

    implements(IDataManager, service=False)

    def __init__(self):
        self._info = []
        self.options = Options()

    def initialize(self, **kwds):
        self.filename = kwds.pop('filename')
        self.add_options(**kwds)

    def add_options(self, **kwds):
        self.options.update(kwds)

    def open(self):
        if self.filename is None:
            raise IOError("No filename specified")
        if not os.path.exists(self.filename):
            raise IOError("Cannot find file '%s'" % self.filename)

    def close(self):
        pass

    def read(self):
        """
        This function does nothing, since executing AMPL data commands
        both reads and processes the data all at once.
        """
        return True

    def write(self, data):
        """
        Return False, because we cannot write to a *.dat file.
        """
        return False

    def process(self, model, data, default):
        """
        Read AMPL data commands and process the data.
        """
        _process_include(['include', self.filename], model, data, default, self.options)

    def clear(self):
        self._info = []



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

__all__ = ['RangeSet']

import math
from six.moves import xrange

from coopr.pyomo.base.sets import SimpleSet
from coopr.pyomo.base.expr import _ExpressionBase
from coopr.pyomo.base.set_types import Integers, Reals
from coopr.pyomo.base.misc import apply_indexed_rule
from coopr.pyomo.base.numvalue import value
from coopr.pyomo.base.component import register_component


class RangeSet(SimpleSet):
    """
    A set that represents a list of numeric values.
    """

    def __init__(self, *args, **kwds):
        """
        Construct a list of integers
        """
        if len(args) == 0:
            raise RuntimeError("Attempting to construct a RangeSet object with no arguments!")
        SimpleSet.__init__(self, **kwds)
        self._type=RangeSet
        #
        if len(args) == 1:
            #
            # RangeSet(end) generates the set: 1 ... end
            #
            self._start=1
            self._end=args[0]
            self._step=1
        elif len(args) == 2:
            #
            # RangeSet(start,end) generates the set: start ... end
            #
            self._start=args[0]
            self._end=args[1]
            self._step=1
        else:
            #
            # RangeSet(start,end,step) generates the set: start, start+step, start+2*step, ... end
            #
            self._start=args[0]
            self._end=args[1]
            self._step=args[2]
        #
        self.ordered = True     # This is an ordered set
        self.value = None       # No internal set data
        self.virtual = True     # This is a virtual set
        self.concrete = True    # This is a concrete set
        self._len = 0           # This is set by the construct() method

    def construct(self, values=None):
        """
        Initialize set data
        """
        if self._constructed:
            return
        self._constructed=True
        #
        # Set the value of the set parameters using
        # expressions
        #
        if isinstance(self._start,_ExpressionBase):
            self._start_val = self._start()
        else:
            self._start_val = value(self._start)
        if isinstance(self._end,_ExpressionBase):
            self._end_val = self._end()
        else:
            self._end_val = value(self._end)
        if isinstance(self._step,_ExpressionBase):
            self._step_val = self._step()
        else:
            self._step_val = value(self._step)
        #
        # The set generates integer values if the starting value, 
        # step and end value are all integers.  Otherwise, the set
        # generates real values.
        #
        if type(self._start_val) is int and type(self._step) is int and type(self._end_val) is int:
            self.domain = Integers
        else:
            self.domain = Reals
        #
        # Compute the set length and upper bound
        #
        if self.filter is None and self.validate is None:
            #
            # Directly compute the number of elements in the set, from 
            # which the upper-bound is computed.
            #
            self._len = int(math.floor((self._end_val-self._start_val+self._step_val+1e-7)//self._step_val))
            ub = self._start_val + (self._len-1)*self._step_val
        else:
            #
            # Iterate through the set to compute the upper bound
            # and number of elements.
            #
            ub = self._start_val
            ctr=0
            for i in self:
                ub = i
                ctr += 1
            self._len = ctr
        #
        # Set the bounds information
        #
        self._bounds = (self._start_val, ub)

    def __len__(self):
        """
        Return the pre-computed set length
        """
        return self._len

    def __iter__(self):
        if not self._constructed:
            raise RuntimeError("A RangeSet component must be initialized before a user can iterate over its elements.")
        if self.filter is None and self.validate is None:
            #
            # Iterate through all set elements
            #
            for i in xrange(self._len):
                yield self._start_val + i*self._step_val
        else:
            #
            # Iterate through all set elements and filter 
            # and/or validate the element values.
            #
            for i in xrange(int((self._end_val-self._start_val+self._step_val+1e-7)//self._step_val)):
                val = self._start_val + i*self._step_val
                if not self.filter is None and not apply_indexed_rule(self, self.filter, self._parent(), val):
                    continue
                if not self.validate is None and not apply_indexed_rule(self, self.validate, self._parent(), val):
                    continue
                yield val

    def data(self):
        """The underlying set data."""
        return set(self)

    def first(self):
        """The first element is the lower bound"""
        return self._bounds[0]

    def last(self):
        """The last element is the upper bound"""
        return self._bounds[1]

    def member(self, key):
        """
        Return the value associated with this key.  Valid
        index values are 1 .. len(set), or -1 .. -len(set).
        Negative key values index from the end of the set.
        """
        if key >= 1:
            if key > self._len:
                raise IndexError("Cannot index a RangeSet past the last element")
            return self._start_val + (key-1)*self._step_val
        elif key < 0:
            if self._len+key < 0:
                raise IndexError("Cannot index a RangeSet past the first element")
            return self._start_val + (self._len+key)*self._step_val
        else:
            raise IndexError("Valid index values for sets are 1 .. len(set) or -1 .. -len(set)")

    def __eq__(self, other):
        """
        Equality comparison

        RangeSet inherits from SimpleSet, so we need to
        replicate this logic from SetOf.__eq__
        """
        if other is None:
            return False
        other = self._set_repn(other)
        if self.dimen != other.dimen:
            return False
        if other.concrete and len(self) != len(other):
            return False
        ctr = 0
        for i in other:
            if not i in self:
                return False
            ctr += 1
        return other.concrete or ctr == len(self)

    def _set_contains(self, element):
        """
        Test if the specified element in this set.
        """
        try:
            x = element - self._start_val
            if x % self._step_val != 0:
                #
                # If we are doing floating-point arithmetic, there is a
                # chance that we are seeing roundoff error...
                #
                if math.fabs((x + 1e-7) % self._step_val) > 2e-7:
                    return False
            if element < self._bounds[0] or element > self._bounds[1]:
                return False
        except:
            #
            # This exception is triggered when type(element) is not int or float.
            #
            return False
        #
        # Now see if the element if filtered or invalid.
        #
        if self.filter is not None and not self.filter(element):
            return False
        if self.validate is not None and not self.validate(self, element):
            return False
        return True


register_component(RangeSet, "A sequence of numeric values.  RangeSet(start,end,step) is a sequence starting a value 'start', and increasing in values by 'step' until a value greater than or equal to 'end' is reached.")




#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2010 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the FAST README.txt file.
#  _________________________________________________________________________

__all__ = ['linearize_model_expressions', 'is_linear_expression_constant']

# a prototype utility to translate from a "normal" model instance containing
# constraints and objectives with expresssions for their bodies, into a
# more compact linearized verson of the problem. eliminates the need for
# canonical expressions and the associated overhead (memory and speed).

# NOTE: Currently only works for constraints - leaving the expressions alone,
#       mainly because PH doesn't modify those and I don't want to deal with
#       that aspect for the moment. In particular, this only works for
#       immutable parameters due to the reliance on the canonical expression
#       generator to extract the linear and constant terms.

from six import iteritems, itervalues
from coopr.pyomo import *
from coopr.pyomo.expr import generate_canonical_repn, canonical_is_linear, canonical_is_constant

def linearize_model_expressions(instance):
    var_id_map = {}

    for block in instance.all_blocks():
        all_repns = list()

        block_repn_suffix = block.subcomponent("canonical_repn")
        if (block_repn_suffix is None) or (not block_repn_suffix.type() is Suffix) or (not block_repn_suffix.active is True):
            block_repn_suffix = None

        # TBD: Should we really be doing this for all components, and not just active ones?
        for constraint_data in active_subcomponents_data_generator(block,Constraint):
            canonical_encoding = None
            if block_repn_suffix is not None:
                canonical_encoding = block_repn_suffix.getValue(constraint_data)
            if canonical_encoding is None:
                canonical_encoding = generate_canonical_repn(constraint_data.body, var_id_map)

            # we obviously can't linearize an expression if it has higher-order terms!
            if canonical_is_linear(canonical_encoding) or canonical_is_constant(canonical_encoding):

                variable_map = canonical_encoding.variables

                constant_term = 0.0
                linear_terms = [] # a list of coefficient, _VarData pairs.

                if canonical_encoding.constant != None:
                    constant_term = canonical_encoding[0]

                for i in xrange(0, len(canonical_encoding.linear)):
                    var_coefficient = canonical_encoding.linear[i]
                    var_value = canonical_encoding.variables[i]
                    linear_terms.append((var_coefficient, var_value))

                # eliminate the expression tree - we don't need it any longer.
                # ditto the canonical representation.
                constraint_data.body = None
                all_repns.append((constraint_data,[constant_term, linear_terms]))
                if block_repn_suffix is not None:
                    block_repn_suffix.clearValue(constraint_data)

        # Just overwrite any existing component that be called lin_body even if it is a component
        block.lin_body = Suffix(direction=Suffix.LOCAL,datatype=None)
        # expand=False because we know there are no array components in this list
        # and this provides a performance improvement
        block.lin_body.updateValues(all_repns, expand=False)

def is_linear_expression_constant(lin_body):
    """Return True if the linear expression is a constant expression, due either to a lack of variables or fixed variables"""
    return (len(lin_body[1]) == 0)



# a simple SOS example from "Modeling Building in Mathematical Programming", H. Paul Williams, 4th Edition, p. 166.

from coopr.pyomo import *

model = AbstractModel()

# the set of customers.
model.Customers = Set()

# the possible locations of the facility that you're trying to place.
model.Sites = Set()

# the cost of satisfying customer demand from each of the potential sites.
model.SatisfactionCost = Param(model.Customers, model.Sites, within=NonNegativeReals)

# indicators of which sites are selected. constraints ensure only one site is selected,
# and allow the binary integrality to be implicit.
model.SiteSelected = Var(model.Sites, bounds=(0,1))

# ensure that only one of the site selected variables is non-zero.
model.SiteSelectedSOS = SOSConstraint(var=model.SiteSelected, sos=1)

# ensure that one of the sites is selected (enforce binary).
def enforce_site_selected_binary_rule(model):
    return summation(model.SiteSelected) == 1
model.EnforceSiteSelectedBinary = Constraint(rule=enforce_site_selected_binary_rule)

# the objective is to minimize the cost to satisfy all customers.
def minimize_cost_rule(model):
    return sum([model.SatisfactionCost[c, s] * model.SiteSelected[s] for c in model.Customers for s in model.Sites])
model.MinimizeCost = Objective(rule=minimize_cost_rule, sense=minimize)




import coopr.pyomo.plugins.data.ampl
import coopr.pyomo.plugins.data.csv_table
import coopr.pyomo.plugins.data.db_table
import coopr.pyomo.plugins.data.json_dict
import coopr.pyomo.plugins.data.sheet
import coopr.pyomo.plugins.data.text
import coopr.pyomo.plugins.data.xml_table




#
# Test the Pyomo command-line interface
#

import unittest
import os
import sys
from os.path import abspath, dirname
currdir = dirname(abspath(__file__))+os.sep
scriptdir = dirname(dirname(dirname(dirname(dirname(abspath(__file__))))))+os.sep
scriptdir = os.path.join(scriptdir,'examples','pyomo','p-median')+os.sep

import re
from six import StringIO
import pyutilib.services
import pyutilib.subprocess
import pyutilib.th as unittest
from pyutilib.misc import setup_redirect, reset_redirect
import coopr.pyomo
import coopr.pyomo.scripting.pyomo as main
import coopr.environ

if os.path.exists(sys.exec_prefix+os.sep+'bin'+os.sep+'coverage'):
    executable=sys.exec_prefix+os.sep+'bin'+os.sep+'coverage -x '
else:
    executable=sys.executable

def filter_fn(line):
    #print line
    tmp = line.strip()
    #print line.startswith('    ')
    return tmp.startswith('Disjunct') or tmp.startswith('DEPRECATION') or tmp.startswith('DiffSet') or line.startswith('    ') or tmp.startswith("Differential") or tmp.startswith("DerivativeVar") or tmp.startswith("InputVar") or tmp.startswith('StateVar') or tmp.startswith('Complementarity')


_diff_tol = 1e-6

@unittest.skipIf(pyutilib.services.registered_executable("glpsol") is None, "The 'glpsol' executable is not available")
class Test(unittest.TestCase):

    def pyomo(self, cmd, **kwds):
        args=re.split('[ ]+',cmd)
        if 'root' in kwds:
            OUTPUT=kwds['root']+'.out'
            results=kwds['root']+'.jsn'
            self.ofile = OUTPUT
        else:
            OUTPUT=StringIO()
            results='results.jsn'
        setup_redirect(OUTPUT)
        os.chdir(currdir)
        output = main.run(['--json', '--save-results=%s' % results] + list(args))
        reset_redirect()
        if not 'root' in kwds:
            return OUTPUT.getvalue()
        return output

    def setUp(self):
        self.ofile = None

    def tearDown(self):
        return
        if self.ofile and os.path.exists(self.ofile):
            return
            os.remove(self.ofile)
        if os.path.exists(currdir+'results.jsn'):
            return
            os.remove(currdir+'results.jsn')

    def run_pyomo(self, cmd, root=None):
        return pyutilib.subprocess.run('pyomo --json --save-results=%s.jsn ' % (root) +cmd, outfile=root+'.out')

    def test1(self):
        """Simple execution of 'pyomo'"""
        self.pyomo(currdir+'pmedian.py pmedian.dat', root=currdir+'test1')
        self.assertMatchesJsonBaseline(currdir+"test1.jsn", currdir+"test1.txt",tolerance=_diff_tol)
        os.remove(currdir+'test1.out')

    def test1a(self):
        """Simple execution of 'pyomo'"""
        self.run_pyomo(currdir+'pmedian.py pmedian.dat', root=currdir+'test1a')
        self.assertMatchesJsonBaseline(currdir+"test1a.jsn", currdir+"test1.txt",tolerance=_diff_tol)
        os.remove(currdir+'test1a.out')

    def test2(self):
        """Run pyomo with bad --model-name option value"""
        self.pyomo('--model-name=dummy pmedian.py pmedian.dat', root=currdir+'test2')
        def filter(line):
            return line.startswith('[') or line.startswith('DEPRECATION')
        self.assertFileEqualsBaseline(currdir+"test2.out", currdir+"test2.txt", filter=filter)

    def test3(self):
        """Run pyomo with model that does not define model object"""
        self.pyomo('pmedian1.py pmedian.dat', root=currdir+'test3')
        def filter(line):
            return line.startswith('[') or line.startswith('DEPRECATION')
        self.assertFileEqualsBaseline(currdir+"test3.out", currdir+"test3.txt", filter=filter)

    def test4(self):
        """Run pyomo with good --model-name option value"""
        self.run_pyomo('-k -l --model-name=MODEL '+currdir+'pmedian1.py pmedian.dat', root=currdir+'test4')
        self.assertMatchesJsonBaseline(currdir+"test4.jsn", currdir+"test1.txt",tolerance=_diff_tol)
        os.remove(currdir+'test4.out')

    def test5(self):
        """Run pyomo with create_model function"""
        self.pyomo('pmedian2.py pmedian.dat', root=currdir+'test5')
        def filter(line):
            return ("Writing model " in line) or ("Solver results file" in line) or \
                   line.startswith('[') or line.startswith('DEPRECATION')
        self.assertFileEqualsBaseline(currdir+"test5.out", currdir+"test5.txt",filter=filter,tolerance=_diff_tol)
        os.remove(currdir+'test5.jsn')

    def test6(self):
        """Run pyomo with help-components option"""
        self.pyomo('--help-components', root=currdir+'test6')
        self.assertFileEqualsBaseline(currdir+"test6.out", currdir+"test6.txt", filter=filter_fn)

    def Xtest7(self):
        """Run pyomo with help option"""
        self.pyomo('--help', root=currdir+'test7')
        self.assertMatchesJsonBaseline(currdir+"test7.jsn", currdir+"test7.txt")

    def test8(self):
        """Run pyomo with --instance-only option"""
        output = self.pyomo('--instance-only pmedian.py pmedian.dat', root=currdir+'test8')
        self.assertEqual(type(output.retval.instance), coopr.pyomo.AbstractModel)
        # Check that the results file was NOT created
        self.assertRaises(OSError, lambda: os.remove(currdir+'test8.jsn'))
        os.remove(currdir+'test8.out')

    def test9(self):
        """Run pyomo with --disable-gc option"""
        output = self.pyomo('--disable-gc pmedian.py pmedian.dat', root=currdir+'test9')
        self.assertEqual(type(output.retval.instance), coopr.pyomo.AbstractModel)
        os.remove(currdir+'test9.jsn')
        os.remove(currdir+'test9.out')

    def Xtest10(self):
        """Run pyomo with --verbose option"""
        def filter(line):
            return ("Writing model " in line) or ("Solver results file" in line) or \
                   line.startswith('DEBUG:') or line.startswith('INFO:') or \
                   line.startswith('[') or line.endswith('cpxlp') or line.startswith('DEPRECATION') or \
                    line.startswith('WARNING')
        self.pyomo('-v pmedian.py pmedian.dat', root=currdir+'test10')
        self.assertFileEqualsBaseline(currdir+"test10.out", currdir+"test10.txt", filter)
        os.remove(currdir+'test10.jsn')

    def Xtest11(self):
        """Run pyomo with --debug=generate option"""
        self.pyomo('--debug=generate pmedian.py pmedian.dat', root=currdir+'test11')
        self.assertFileEqualsBaseline(currdir+"test11.jsn", currdir+"test11.txt")

    def test12(self):
        """Run pyomo with --output option"""
        def filter(line):
            return ("Writing model " in line) or line.startswith('DEPRECATION')
        self.pyomo('--output=%s pmedian.py pmedian.dat' % (currdir+'test12.log'), root=currdir+'test12')
        self.assertMatchesJsonBaseline(currdir+"test12.jsn", currdir+"test12.txt",tolerance=_diff_tol)
        os.remove(currdir+'test12.log')
        os.remove(currdir+'test12.out')

    def test13(self):
        """Simple execution of 'pyomo' with implicit rules"""
        self.pyomo('pmedian3.py pmedian.dat', root=currdir+'test13')
        self.assertMatchesJsonBaseline(currdir+"test13.jsn", currdir+"test1.txt",tolerance=_diff_tol)
        os.remove(currdir+'test13.out')

    def test14(self):
        """Simple execution of 'pyomo' with a concrete model and constraint lists"""
        self.pyomo('pmedian4.py', root=currdir+'test14')
        self.assertMatchesJsonBaseline(currdir+"test14.jsn", currdir+"test14.txt",tolerance=_diff_tol)
        os.remove(currdir+'test14.out')


if __name__ == "__main__":
    unittest.main()



# a slightly less brain-dead separable parabolic function of three variables, whose
# minimum is obviously at x[1]=x[2]=x[3]=0.

from coopr.pyomo import *

model = AbstractModel()

def indices_rule(model):
    return xrange(1,4)
model.indices = Set(initialize=indices_rule, within=PositiveIntegers)

model.x = Var(model.indices, within=Reals)

def bound_x_rule(model, i):
    return (-10, model.x[i], 10)
model.bound_x = Constraint(model.indices, rule=bound_x_rule)

def objective_rule(model):
    return sum([model.x[i] * model.x[i] for i in model.indices])
model.objective = Objective(rule=objective_rule, sense=minimize)



#
# Imports
#
from coopr.pyomo import *

#
# Setup
#

model = AbstractModel()

model.PROD = Set()

model.rate = Param(model.PROD,within=PositiveReals)

model.avail = Param(within=NonNegativeReals)

model.profit = Param(model.PROD)

model.market = Param(model.PROD, within=NonNegativeReals)

def Make_bounds(model,i):
    return (0,model.market[i])
model.Make = Var(model.PROD, bounds=Make_bounds)

def Objective_rule(model):
    return summation(model.profit, model.Make)
model.Total_Profit = Objective(rule=Objective_rule, sense=maximize)

def Time_rule(model):
    ans = 0
    for p in model.PROD:
        ans = ans + (1.0/model.rate[p]) * model.Make[p]
    return ans < model.avail

def XTime_rule(model):
    return summation(model.Make, denom=(model.rate,) ) < model.avail
#model.Time = Constraint(rule=Time_rule)



#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

__all__ = ['Block',
           'active_subcomponents_generator',
           'subcomponents_generator',
           'active_subcomponents_data_generator',
           'subcomponents_data_generator']

import sys
import weakref
import logging
from operator import itemgetter
from six import iterkeys, iteritems, itervalues, StringIO

from pyutilib.misc import Container

from coopr.pyomo.base.plugin import *
from coopr.pyomo.base.component import Component, ActiveComponentData, ComponentUID
from coopr.pyomo.base.sets import Set, SimpleSet
from coopr.pyomo.base.var import Var
from coopr.pyomo.base.misc import apply_indexed_rule
from coopr.pyomo.base.sparse_indexed_component import SparseIndexedComponent, ActiveSparseIndexedComponent

logger = logging.getLogger('coopr.pyomo')


# 
#
#

class _BlockConstruction(object):
    """
    This class holds a "global" dict used when constructing
    (hierarchical) models.
    """
    data = {}


class _BlockData(ActiveComponentData):
    """
    This class holds the fundamental block data.
    """

    class PseudoMap(object):
        """
        This class presents a "mock" dict interface to the internal
        BlockComponents data structures.  We return this object to the
        user to preserve the historical "{ctype : {name : obj}}"
        interface without actually regenerating that dict-of-dicts data
        structure.
        """

        __slots__ = ( '_block', '_ctype', '_active' )
        
        def __init__(self, block, ctype, active=None):
            self._block = block
            self._ctype = ctype
            self._active = active

        def __iter__(self):
            return self.iterkeys()

        def __getitem__(self, key):
            if key in self._block._decl:
                x = self._block._decl_order[self._block._decl[key]]
                if self._ctype is None or x[0].type() is self._ctype:
                    if self._active is None or x[0].active == self._active:
                        return x[0]
            msg = ""
            if self._active is not None:
                msg += self._active and "active " or "inactive "
            if self._ctype is not None:
                msg += self._ctype.__name__ + " "
            raise KeyError( "%scomponent %s not found in block %s" 
                            % (msg, key, self._block.cname()))

        def __nonzero__(self):
            # Shortcut: this will bail after finding the first item
            for x in itervalues(self):
                return True
            return False

        __bool__ = __nonzero__

        def __len__(self):
            if self._active is None:
                if self._ctype is None:
                    return sum(x[2] for x in itervalues(self._block._ctypes))
                else:
                    return self._block._ctypes.get(self._ctype,(0,0,0))[2]
            #For active==True/False, we have to count by brute force
            ans = 0
            for x in itervalues(self):
                ans += 1
            return ans

        def __contains__(self, key):
            if key in self._block._decl:
                x = self._block._decl_order[self._block._decl[key]]
                if x[0].type() is self._ctype:
                    if self._active is None or x[0].active == self._active:
                        return True
            return False

        def iterkeys(self):
            if self._ctype is None:
                idx = 0
            else:
                idx = self._block._ctypes.get(self._ctype, [None])[0]
            while idx is not None:
                try:
                    x = self._block._decl_order[idx]
                except IndexError:
                    break

                if x[0] is not None:
                    if self._active is None or x[0].active == self._active:
                        yield x[0].name
                if self._ctype is None:
                    idx += 1
                else:
                    idx = x[1]

        def itervalues(self):
            if self._ctype is None:
                idx = 0
            else:
                idx = self._block._ctypes.get(self._ctype, [None])[0]
            while idx is not None:
                try:
                    x = self._block._decl_order[idx]
                except IndexError:
                    break

                if x[0] is not None:
                    if self._active is None or x[0].active == self._active:
                        yield x[0]
                if self._ctype is None:
                    idx += 1
                else:
                    idx = x[1]

        def iteritems(self):
            if self._ctype is None:
                idx = 0
            else:
                idx = self._block._ctypes.get(self._ctype, [None])[0]
            while idx is not None:
                try:
                    x = self._block._decl_order[idx]
                except IndexError:
                    break

                if x[0] is not None:
                    if self._active is None or x[0].active == self._active:
                        yield (x[0].name, x[0])
                if self._ctype is None:
                    idx += 1
                else:
                    idx = x[1]

        def keys(self):
            return list(self.iterkeys())

        def values(self):
            return list(self.itervalues())

        def items(self):
            return list(self.iteritems())

    def __init__(self, owner):
        #
        # We used to define an internal structure that looked like:
        #
        #    _component    = { ctype : OrderedDict( name : obj ) }
        #    _declarations = OrderedDict( name : obj )
        #
        # This structure is convenient, but the overhead of carrying
        # around roughly 20 dictionaries for every block consumed a
        # nontrivial amount of memory.  Plus, the generation and
        # maintenance of OrderedDicts appeared to be disturbingly slow.
        #
        # We now "mock up" this data structure using 2 dicts and a list:
        #
        #    _ctypes     = { ctype : [ first idx, last idx, count ] }
        #    _decl       = { name  : idx }
        #    _decl_order = [ (obj, next_ctype_idx) ]
        #
        # Some notes: As Pyomo models rarely *delete* objects, we
        # currently never remove items from the _decl_order list.  If
        # the component is ever removed / cleared, we simply mark the
        # object as None.  If models crop up where we start seeing a
        # significant amount of adding / removing components, then we
        # can revisit this decision (although we will probably continue
        # marking entries as None and just periodically rebuild the list
        # as opposed to maintaining the list without any holes.
        #
        ActiveComponentData.__init__(self, owner)
        # Note: call super() here to bypass the Block __setattr__
        #   _ctypes:      { ctype -> [1st idx, last idx, count] }
        #   _decl:        { name -> idx }
        #   _decl_order:  list( tuples( obj, next_type_idx ) )
        super(_BlockData, self).__setattr__('_ctypes', {})
        super(_BlockData, self).__setattr__('_decl', {})
        super(_BlockData, self).__setattr__('_decl_order', [])


    def __getstate__(self):
        # Note: _BlockData is NOT slot-ized, so we must pickle the
        # entire __dict__.  However, we want the base class's
        # __getstate__ to override our blanket approach here (i.e., it
        # will handle the _component weakref).
        ans =  dict(self.__dict__)
        ans.update(super(_BlockData, self).__getstate__())
        return ans

    def __setstate__(self, state):
        # We want the base class's __setstate__ to override our blanket
        # approach here (i.e., it will handle the _component weakref).
        for (slot_name, value) in iteritems(state):
            super(_BlockData, self).__setattr__(slot_name, value)
        super(_BlockData, self).__setstate__(state)


    #def __getitem__(self, name):
    #    # This supports two key functionalities:
    #    #  1. block["component_name"] as a convenient alternative to
    #    #     block.component_name
    #    #  2. block[Var] as an alternative to block.subcomponents(Var)
    #
    #    if name in self._decl:
    #        return self._decl_order[self._decl[name]][0]
    #    elif name in self._ctypes:
    #        return BlockComponents.PseudoMap(self, name)
    #    elif isinstance(name,basestring):
    #        return None
    #    else:
    #        return {}

    def __setattr__(self, name, val):
        #
        # Set an attribute on this Block.  In general, the most common
        # case for this is setting a *new* attribute.  After that, there
        # is updating an existing Component value, with the least common
        # case being resetting an existing general attribute.  Given
        # that, we will:
        #
        # 1) handle new attributes, with special handling of Components
        if name not in self.__dict__:
            if isinstance(val, Component):
                self.add_component(name, val)
            else:
                super(_BlockData, self).__setattr__(name, val)
                #self.__dict__[name] = val
        # 2) Update the value of existing [scalar] Components
        # (through the component's set_value())
        elif name in self._decl:
            # JDS: setting a component to None should attempt
            #if val is None:
            #    self.del_component(name)
            if isinstance(val, Component):
                logger.warn(
                    "Attempting to implictly replace the Component attribute "
                    "%s (type=%s) on block %s with a new Component (type=%s)."
                    "\nThis is usually indicative of a modelling error.\n"
                    "To avoid this warning, use block.del_component() and "
                    "block.add_component()."
                    % ( name, type(self.subcomponent(name)), self.cname(True), 
                        type(val) ) )
                self.del_component(name)
                self.add_component(name, val)
            else:
                try:
                   self.subcomponent(name).set_value(val)
                except AttributeError:
                    logger.error(
                        "Expected component %s (type=%s) on block %s to have a "
                        "'set_value' method, but none was found." % 
                        ( name, type(self.subcomponent(name)), 
                          self.cname(True) ) )
                    raise
        # 3) handle setting non-Component attributes
        else:
            # NB: This is important: the _BlockData is either a
            # scalar Block (where _parent and _component are defined)
            # or a single block within an Indexed Block (where only
            # _component is defined).  Regardless, the __init__()
            # declares these attributes and sets them to either None or
            # a weakref.  This means that we will never have a problem
            # converting these objects from weakrefs into Blocks and
            # back (when pickling): because the attribute is already
            # in __dict__, we will not hit the add_component /
            # del_component branches above.  It also means that any
            # error checking we want to do when assigning these
            # attributes should be done here.

            # NB: isintance() can be slow; however, it is only slow when
            # it returns False.  Since the common paths on this branch
            # should return True, this shouldn't be too inefficient.
            if name == '_parent':
                if val is not None and not isinstance(val(), _BlockData):
                    raise ValueError(
                        "Cannot set the '_parent' attribute of Block '%s' "
                        "to a non-Block object (with type=%s); Did you "
                        "try to create a model component named '_parent'?" 
                        % (self.cname(), type(val)) )
                super(_BlockData, self).__setattr__(name, val)
                #self.__dict__[name] = val
            elif name == '_component':
                if val is not None and not isinstance(val(), _BlockData):
                    raise ValueError(
                        "Cannot set the '_component' attribute of Block '%s' "
                        "to a non-Block object (with type=%s); Did you "
                        "try to create a model component named '_component'?" 
                        % (self.cname(), type(val)) )
                super(_BlockData, self).__setattr__(name, val)
                #self.__dict__[name] = val
            # At this point, we should only be seeing non-component data
            # the user is hanging on the blocks (uncommon) or the
            # initial setup of the object data (in __init__).
            elif isinstance(val, Component):
                logger.warn(
                    "Attempting to reassign the non-component attribute %s "
                    "on block %s with a new Component with type %s.  This is "
                    "usually indicative of a modelling error."
                    % (name, self.cname(True), type(val)) )
                self.del_component(name)
                self.add_component(name, val)
            else:
                super(_BlockData, self).__setattr__(name, val)
                #self.__dict__[name] = val
 
    def __delattr__(self, name):
        # We need to make sure that del_component() gets called if a
        # user attempts to remove an attribute from a block that happens
        # to be a component
        if name in self._decl:
            self.del_component(name)
        else:
            super(_BlockData, self).__delattr__(name)

    def _add_temporary_set(self,val):
        """ TODO """
        _component_sets = getattr(val, '_implicit_subsets', None)
        #
        # FIXME: The name attribute should begin with "_", and None
        # should replace "_unknown_"
        #
        if _component_sets is not None:
            ctr=0
            for tset in _component_sets:
                if tset.cname() == "_unknown_":
                    self._construct_temporary_set(
                      tset,
                      val.cname()+"_index_"+str(ctr)
                    )
                ctr += 1
        if isinstance(val._index,SimpleSet) and val._index.name == "_unknown_":
            self._construct_temporary_set(val._index,val.cname()+"_index")
        if getattr(val,'domain',None) is not None and val.domain.name == "_unknown_":
            self._construct_temporary_set(val.domain,val.cname()+"_domain")

    def _construct_temporary_set(self, obj, name):
        """ TODO """
        if type(obj) is tuple:
            if len(obj) == 1:                #pragma:nocover
                raise Exception(
                    "Unexpected temporary set construction for set "
                    "%s on block %s" % ( name, self.cname()) )
            else:
                tobj = obj[0]
                for t in obj[1:]:
                    tobj = tobj*t
                setattr(self,name,tobj)
                tobj.virtual=True
                return tobj
        if isinstance(obj,Set):
            setattr(self,name,obj)
            return obj

    def add_component(self, name, val):
        if not val.valid_model_component():
            raise RuntimeError(
                "Cannot add '%s' as a component to a model" % str(type(val)) )
        if name in self.__dict__:
            raise RuntimeError(
                "Cannot add component '%s' (type %s) to block '%s': a "
                "component by that name (type %s) is already defined."
                % (name, type(val), self.cname(), type(getattr(self, name))))

        _component = self.component()
        _type = val.type()
        if _type in _component._suppress_ctypes:
            return

        # all Pyomo components have Parents
        if (val._parent is not None) and (val._parent() is not None):
            if val._parent() is self:
                msg = """
Attempting to re-assign the component '%s' to the same 
block under a different name (%s).""" % (val.name, name)
            else:
                msg = """
Re-assigning the component '%s' from block '%s' to 
block '%s' as '%s'.""" % ( val.name, val._parent().cname(True), 
                           self.cname(True), name )

            raise RuntimeError("""%s

This behavior is not supported by Pyomo: components must have a single
owning block (or model) and a component may not appear multiple times in
a block.  If you want to re-name or move this component, use the block
del_component() and add_component() methods.  We are cowardly refusing
to do this automatically as renaming the component will change its
position in the construction order for Abstract Models; potentially
leading to unintuitive data validation and construction errors.
""" % (msg.strip(),) ) 

        # all Pyomo components have names. 
        val.name = name
        val._parent = weakref.ref(self)

        # We want to add the temporary / implicit sets first so that
        # they get constructed before this component
        self._add_temporary_set(val)

        # Add the component to the underlying Component store
        _new_idx = len(self._decl_order)
        self._decl[name] = _new_idx
        self._decl_order.append( (val, None) )

        # Add the component as an attribute
        super(_BlockData, self).__setattr__(name, val)
        #self.__dict__[name]=val

        # Update the ctype linked lists
        if _type in self._ctypes:
            idx_info = self._ctypes[_type]
            tmp = idx_info[1]
            self._decl_order[tmp] = (self._decl_order[tmp][0], _new_idx)
            idx_info[1] = _new_idx
            idx_info[2] += 1
        else:
            self._ctypes[_type] = [_new_idx, _new_idx, 1]

        # There are some properties that need to be propagated to sub-blicks:
        if _type is Block:
            val._suppress_ctypes |= _component._suppress_ctypes

        # Support implicit rule names
        if '_rule' in val.__dict__:
            if val._rule is None:
                frame = sys._getframe(2)
                locals_ = frame.f_locals
                if val.cname()+'_rule' in locals_:
                    val._rule = locals_[val.cname()+'_rule']
        # FIXME: This is a HACK to support the way old Blocks and legacy
        # IndexedComponents (like Set) behave.  In particular, Set does
        # not define a "rule" attribute.  I put the hack back in to get
        # some tests passing again, but in all honesty, I am amazed this
        # ever worked properly. [JDS]
        elif getattr(val, 'rule', None) is None:
            frame = sys._getframe(2)
            locals_ = frame.f_locals
            if val.cname()+'_rule' in locals_:
                val.rule = locals_[val.cname()+'_rule']

        # Don't reconstruct if this component has already been constructed,
        # the user may just want to transer it to another parent component
        if val._constructed is True:
            return

        # If the block is Concrete, construct the component
        # Note: we are explicitly using getattr because (Scalar)
        #   classes that derive from Block may want to declare components
        #   within their __init__() [notably, coopr.gdp's Disjunct).
        #   Those components are added *before* the _constructed flag is
        #   added to the class by Block.__init__()
        if getattr(_component, '_constructed', False):
            # NB: we don't have to construct the temporary / implicit
            # sets here: if necessary, that happens when
            # _add_temporary_set() calls add_component().
            if id(self) in _BlockConstruction.data:
                data = _BlockConstruction.data[id(self)].get(name,None)
            else:
                data = None
            if __debug__ and logger.isEnabledFor(logging.DEBUG):
                # This is tricky: If we are in the middle of
                # constructing an indexed block, the block component
                # already has _constructed=True.  Now, if the
                # _BlockData.__init__() defines any local variables
                # (like coopr.gdp.Disjunct's indicator_var), cname(True)
                # will fail: this block data exists and has a parent(),
                # but it has not yet been added to the parent's _data
                # (so the idx lookup will fail in cname()).
                if self.parent() is None:
                    _blockName = "[Model]"
                else:
                    try:
                        _blockName = "Block '%s'" % self.cname(True)
                    except:
                        _blockName = "Block '%s[...]'" \
                            % self.component().cname(True)
                logger.debug( "Constructing %s '%s' on %s from data=%s",
                              val.__class__.__name__, val.cname(), 
                              _blockName, str(data) )
            try:
                val.construct(data)
            except:
                err = sys.exc_info()[1]
                logger.error(
                    "Constructing component '%s' from data=%s failed:\n%s: %s",
                    str(val.cname(True)), str(data).strip(),
                    type(err).__name__, err )
                raise
            if __debug__ and logger.isEnabledFor(logging.DEBUG):
                if _blockName[-1] == "'":
                    _blockName = _blockName[:-1] + '.' + val.cname() + "'"
                else:
                    _blockName = "'" + _blockName + '.' + val.cname() + "'"
                _out = StringIO()
                val.pprint(ostream=_out)
                logger.debug( "Constructed component '%s':\n%s" 
                              % ( _blockName, _out.getvalue() ) )

    def del_component(self, name):
        # FIXME: Is this necessary?  Should this raise an exception?
        if name not in self._decl:
            return

        # Remove the attribute
        del self.__dict__[name]

        # Replace the component in the master list
        idx = self._decl[name]
        del self._decl[name]
        obj = self._decl_order[idx][0]
        self._decl_order[idx] = (None, self._decl_order[idx][1])

        # Update the ctype linked lists
        ctype_info = self._ctypes[obj.type()]
        ctype_info[2] -= 1
        if ctype_info[2] == 0:
            del self._ctypes[obj.type()]

        # Clear the _parent attribute
        obj._parent = None

    def reclassify_subcomponent_type(self, name_or_object, new_ctype):
        try:
            c = name_or_object.component()
        except AttributeError:
            c = self.subcomponent(name_or_object)
        if c._type is not new_ctype:
            self.del_component(c.cname())
            c._type = new_ctype
            self.add_component(c.cname(), c)


    #def keys(self):
    #    return list(self.iterkeys())

    #def __iter__(self):
    #    for x in self._decl_order:
    #        if x[0] is not None:
    #            yield x[0].cname()

    #def iterkeys(self):
    #    for x in self._decl_order:
    #        if x[0] is not None:
    #            yield x[0].cname()

    #def itervalues(self):
    #    for x in self._decl_order:
    #        if x[0] is not None:
    #            yield x[0]

    #def iteritems(self):
    #    for x in self._decl_order:
    #        if x[0] is not None:
    #            yield (x[0].cname(), x[0])

    def contains_component(self, ctype):
        return ctype in self._ctypes and self._ctypes[ctype][2]

    def subcomponent_map(self, ctype=None, active=None):
        """
        Return information about the block components.  If ctype is
        None, return the dictionary that maps
           {component type -> {name -> instance}}
        Otherwise, return the dictionary that maps
           {name -> instance}
        for the specified component type.

        Note: the actual {name->instance} object is a PseudoMap that
        implements a lightweight interface to the underlying
        BlockComponents data structures.
        """

        if ctype is None:
            ans = {}
            for x in self._ctypes:
                ans[x] = _BlockData.PseudoMap(self, x, active)
            return ans
        else:
            return _BlockData.PseudoMap(self, ctype, active)

    def components(self, ctype=None, active=None):
        #logger.warn("DEPRECATION WARNING: Block.components() has been "
        #            "renamed Block.subcomponents().")
        return self.subcomponents(ctype, active)

    def active_components(self, ctype=None):
        #logger.warn("DEPRECATION WARNING: Block.active_components() has been "
        #            "renamed Block.active_subcomponents().")
        return self.active_subcomponents(ctype)

    def subcomponent(self, name):
        if name in self._decl:
            return self._decl_order[self._decl[name]][0]
        else:
            return None

    def subcomponents(self, ctype=None, active=None):
        """
        Returns the components in this block as a PseudoMap.  Callers should
        treat the return value as a normal dictionary.
        """
        return _BlockData.PseudoMap(self, ctype, active)

    def subcomponent_data(self, ctype=None, active=None):
        """
        Generator that returns a 3-tuple of (component name, index value,
        and _ComponentData) for every component data in the model
        """
        for name, comp in _BlockData.PseudoMap(self, ctype, active):
            for idx, compData in comp.iteritems():
                if active is None or comp.active == active:
                    yield(name, idx, compData)

    def active_subcomponents(self, ctype=None):
        """
        Returns the active components in this block.  Return values
        match those of .subcomponents().
        """
        return self.subcomponents(ctype, True)

    def active_subcomponent_data(self, ctype=None):
        """
        Generator that returns a 3-tuple of (component name, index value,
        and _ComponentData) for every active component data in the model
        """
        return self.subcomponent_data(ctype, True)

    def fix_all_vars(self):
        for var in itervalues(self.subcomponents(Var)):
            var.fix()
        for block in itervalues(self.subcomponents(Block)):
            block.fix_all_vars()

    def unfix_all_vars(self):
        for var in itervalues(self.subcomponents(Var)):
            var.unfix()
        for block in itervalues(self.subcomponents(Block)):
            block.unfix_all_vars()

    def is_constructed(self):
        """
        A boolean indicating whether or not all *active* components of the
        input model have been properly constructed.
        """
        if not self._constructed:
            return False
        for x in self._decl_order:
            if x[0] is not None and x[0].active and not x[0].is_constructed():
                return False
        return True

    def pprint(self, ostream=None, verbose=False, prefix=""):
        """
        Print a summary of the block info
        """
        if ostream is None:
            ostream = sys.stdout
        #
        # We hard-code the order of the core Pyomo modeling
        # components, to ensure that the output follows the logical order
        # that expected by a user.
        #
        import coopr.pyomo.base.component_order
        items = coopr.pyomo.base.component_order.items + [Block]
        #
        # Collect other model components that are registered
        # with the IModelComponent extension point.  These are appended
        # to the end of the list of the list.
        #
        dynamic_items = set()
        for item in [ModelComponentFactory.get_class(name).component for name in ModelComponentFactory.services()]:
            if not item in items:
                dynamic_items.add(item)
        # extra items get added alphabetically (so output is consistent)
        items.extend(sorted(dynamic_items, key=lambda x: x.__name__))

        for item in items:
            keys = sorted(self.subcomponents(item))
            if not keys:
                continue
            #
            # NOTE: these conditional checks should not be hard-coded.
            #
            ostream.write( "%s%d %s Declarations\n" 
                           % (prefix, len(keys), item.__name__) )
            for key in keys:
                self.subcomponent(key).pprint(
                    ostream=ostream, verbose=verbose, prefix=prefix+'    ' )
            ostream.write("\n")
        #
        # Model Order
        #
        decl_order_keys = self.subcomponents().keys()
        ostream.write("%s%d Declarations: %s\n" 
                      % ( prefix, len(decl_order_keys),
                          ' '.join(str(x) for x in decl_order_keys) ))


class Block(ActiveSparseIndexedComponent):
    """
    Blocks are indexed components that contain other components
    (including blocks).  Blocks have a global attribute that defines
    whether construction is deferred.  This applies to all components
    that they contain except blocks.  Blocks contained by other
    blocks use their local attribute to determine whether construction
    is deferred.

    NOTE: Blocks do not currently maintain statistics about the sets,
    parameters, constraints, etc that they contain, including these
    components from subblocks.
    """

    def __new__(cls, *args, **kwds):
        if cls != Block:
            return super(Block, cls).__new__(cls)
        if args == ():
            return SimpleBlock.__new__(SimpleBlock)
        else:
            return IndexedBlock.__new__(IndexedBlock)

    def __init__(self, *args, **kwargs):
        """Constructor"""
        self._suppress_ctypes = set()
        self._rule = kwargs.pop('rule', None )
        self._options = kwargs.pop('options', None )
        kwargs.setdefault('ctype', Block)
        SparseIndexedComponent.__init__(self, *args, **kwargs)

    def _default(self, idx):
        return self._data.setdefault(idx, _BlockData(self))

    def concrete_mode(self):
        """Configure block to immediately construct components"""
        self.construct()

    def symbolic_mode(self):
        """Configure block to defer construction of components"""
        if self._constructed:
            logger.error(
                "Attempting to return a Concrete model to Symbolic (Abstract) "
                "mode: you will likely experience unexpected and potentially "
                "erroneous behavior.")
        self._constructed=False

    def all_blocks(self, filter_active=True, sort_by_keys=False, sort_by_names=False):
        if filter_active and not self.active:
            return
        blocks = iteritems(self)
        if sort_by_keys:
            blocks = sorted(blocks, key=itemgetter(0))

        for idx, b in blocks:
            if filter_active and not b.active:
                continue
            yield b
            for subblock in subcomponents_generator( b, Block, sort_by_names=sort_by_names ):
                for tmp in subblock.all_blocks( filter_active, sort_by_keys, sort_by_names ):
                    yield tmp

    # flags *all* active variables (on active blocks) and their composite _VarData objects
    # as stale.  fixed variables are flagged as non-stale. the state of
    # all inactive variables is left unchanged.
    def flag_vars_as_stale(self):
        for block in self.all_blocks():
            for variable in active_subcomponents_generator(block,Var):
                variable.flag_as_stale()

    def find_component(self, label_or_component):
        return ComponentUID(label_or_component).find_component_on(self)

    def _old_find_component(self, label):
        """ TODO """
        cList = label.split('.')
        obj = self
        while cList:
            c = cList.pop(0).split(':')
            if len(c) == 1 and c[0].endswith(']') and c[0].count('[') == 1:
                c = c[0][:-1].split('[')
            if len(c) > 1:
                idx = c[1].split(',')
                for i, val in enumerate(idx):
                    if val[0]=='#':
                        idx[i] = int(val[1:])
                    elif val[0]=='$':
                        idx[i] = val[1:]
                    elif val[0]=='!':
                        idx[i] = None
                    else:
                        # last-ditch effort to find things
                        try:
                            tmp = int(val)
                            idx[i] = tmp
                        except ValueError:
                            pass
                        #raise ValueError(
                        #    "missing index type specifier for index %s in canonical label %s" % (val,label) )
                if len(idx) > 1:
                    idx = tuple(idx)
                else:
                    idx = idx[0]
            else:
                idx = None
            try:
                if idx is None:
                    obj = getattr(obj, c[0])
                else:
                    obj = getattr(obj, c[0])[idx]
            except AttributeError:
                return None
        return obj

    def construct(self, data=None):
        """ TODO """
        if __debug__ and logger.isEnabledFor(logging.DEBUG):
            logger.debug( "Constructing %s '%s', from data=%s",
                          self.__class__.__name__, self.cname(), str(data) )
        if self._constructed:
            return
        self._constructed = True

        # We must check that any pre-existing components are
        # constructed.  This catches the case where someone is building
        # a Concrete model by building (potentially pseudo-abstract)
        # sub-blocks and then adding them to a Concrete model block.
        for idx in self._data:
            _block = self[idx]
            for name, obj in iteritems( _block.subcomponents() ):
                if not obj._constructed:
                    if data is None:
                        _data = None
                    else:
                        _data = data.get(name, None)
                    obj.construct(_data)
        
        if self._rule is None:
            return
        # If we have a rule, fire the rule for all indices.  
        # Notes:
        #  - Since this block is now concrete, any components added to
        #    it will be immediately constructed by
        #    block.add_component().
        #  - Since the rule does not pass any "data" on, we build a
        #    scalar "stack" of pointers to block data
        #    (_BlockConstruction.data) that the individual blocks'
        #    add_component() can refer back to to handle subcomponent
        #    construction.
        for idx in self._index:
            _block = self[idx]
            if data is not None and idx in data:
                _BlockConstruction.data[id(_block)] = data[idx]
            obj = apply_indexed_rule(
                None, self._rule, _block, idx, self._options )
            if id(_block) in _BlockConstruction.data:
                del _BlockConstruction.data[id(_block)]

            # TBD: Should we allow skipping Blocks???
            #if obj is Block.Skip and idx is not None:
            #   del self._data[idx]

    def pprint(self, ostream=None, verbose=False, prefix=""):
        if ostream is None:
            ostream = sys.stdout
        subblock = (self._parent is not None) and (self.parent() is not None)
        for key in sorted(self):
            b = self[key]
            if subblock:
                ostream.write(prefix+b.cname(True)+" : Active=%s\n" % str(self.active))
            _BlockData.pprint( b, ostream=ostream, verbose=verbose, 
                               prefix=prefix+'    ' if subblock else prefix )

    def display(self, filename=None, ostream=None):
        """
        Print the Pyomo model in a verbose format.
        """
        if filename is not None:
            OUTPUT=open(filename,"w")
            self.display(ostream=OUTPUT)
            OUTPUT.close()
            return
        if ostream is None:
            ostream = sys.stdout
        if (self._parent is not None) and (self._parent() is not None):
            ostream.write("Block "+self.cname()+'\n')
        else:
            ostream.write("Model "+self.cname()+'\n')
        #
        import coopr.pyomo.base.component_order
        for item in coopr.pyomo.base.component_order.display_items:
            #
            ostream.write("\n")
            ostream.write("  %s:\n" % coopr.pyomo.base.component_order.display_name[item])
            ACTIVE = self.active_subcomponents(item)
            if not ACTIVE:
                ostream.write("    None\n")
            else:
                for obj in itervalues(ACTIVE):
                    obj.display(prefix="    ",ostream=ostream)


class SimpleBlock(_BlockData, Block):

    def __init__(self, *args, **kwds):
        _BlockData.__init__(self, self)
        Block.__init__(self, *args, **kwds)

    def pprint(self, ostream=None, verbose=False, prefix=""):
        Block.pprint(self, ostream, verbose, prefix)


class IndexedBlock(Block):

    def __init__(self, *args, **kwds):
        Block.__init__(self, *args, **kwds)


#register_component(Block, "Blocks are indexed components that contain "
#                   "one or more other model components.")


#
# We now include a number of generators for iterating over various
# components within a block.  the primary purpose is to streamline code,
# as this kind of iteration is done *everywhere* in Coopr. any generally
# useful iteration schemes should be moved into the following segment.
# 

# iterates over all active subcomponents (e.g., Constraint) of the specified type
# in the input block by declaration order
def active_subcomponents_generator(block, ctype, sort_by_names=False):
    if issubclass(ctype, Component):
        if sort_by_names is False:
            return ( subcomponent for subcomponent in itervalues(block.active_subcomponents(ctype)) \
                         if subcomponent.active )
        else:
            return ( subcomponent for name,subcomponent in sorted(iteritems(block.active_subcomponents(ctype)), key=itemgetter(0)) \
                         if subcomponent.active )
    raise TypeError("The ctype %s is not a subclass of %s" % (ctype, Component))

# iterates over all subcomponents (e.g., Var) of the specified type
# in the input block by declaration order
def subcomponents_generator(block, ctype, sort_by_names=False):
    if issubclass(ctype, Component):
        if sort_by_names is False:
            return ( subcomponent for subcomponent in itervalues(block.subcomponents(ctype)) )
        else:
            return ( subcomponent for name,subcomponent in sorted(iteritems(block.subcomponents(ctype)), key=itemgetter(0)) )
    raise TypeError("The ctype %s is not a subclass of %s" % (ctype, Component))

# iterates over all active subcomponents_datas (e.g., _VarData) of all active
# subcomponents of the specified type in the input block.  If sort_by_names
# is set to True, the objects will be returned in a deterministic order sorted by name
# otherwise objects will be return by declartion order. Within each subcomponent, data objects
# are returned in arbitrary (non-determinisitic) order, unless sort_by_keys is set to True
def active_subcomponents_data_generator(block, ctype, sort_by_keys=False, sort_by_names=False):
    active_subcomponents = \
        active_subcomponents_generator( block, ctype, sort_by_names=sort_by_names)
    if issubclass(ctype, ActiveSparseIndexedComponent):
        for subcomponent in active_subcomponents:
            subCompDataList = iteritems(subcomponent)
            if sort_by_keys:
                subCompDataList = sorted(subCompDataList, key=itemgetter(0))
            for index, subcomponentData in subCompDataList:
                if subcomponentData.active:
                    yield subcomponentData
    elif issubclass(ctype, SparseIndexedComponent):
        for subcomponent in active_subcomponents:
            subCompDataList = iteritems(subcomponent)
            if sort_by_keys:
                subCompDataList = sorted(subCompDataList, key=itemgetter(0))
            for index, subcomponentData in subCompDataList:
                yield subcomponentData
    #### HACK FOR SET (They are not SparseIndexedComponents)
    elif issubclass(ctype, Set):
        raise NotImplementedError("This generator has not been tested for this ctype %s" % ctype)
    ####
    elif issubclass(ctype, Component):
        for subcomponentData in active_subcomponents:
            yield subcomponentData
    else:
        raise TypeError("The ctype %s is not a subclass of %s" % (ctype, Component))

#
# Same as above but don't check the .active flag
#

def subcomponents_data_generator(block, ctype, sort_by_keys=False, sort_by_names=False):
    subcomponents = active_subcomponents_generator( block, ctype, sort_by_names=sort_by_names)
    if issubclass(ctype, SparseIndexedComponent):
        for subcomponent in subcomponents:
            subCompDataList = iteritems(subcomponent)
            if sort_by_keys:
                subCompDataList = sorted(subCompDataList, key=itemgetter(0))
            for index, subcomponentData in subCompDataList:
                yield subcomponentData
    #### HACK FOR SET (They are not SparseIndexedComponents)
    elif issubclass(ctype, Set):
        raise NotImplementedError("This generator has not been tested for this ctype %s" % ctype)
    ####
    elif issubclass(ctype, Component):
        for subcomponentData in subcomponents:
            yield subcomponentData
    else:
        raise TypeError("The ctype %s is not a subclass of %s" % (ctype, Component))




#  _________________________________________________________________________
#
#  Coopr: A COmmon Optimization Python Repository
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  For more information, see the Coopr README.txt file.
#  _________________________________________________________________________

import coopr.pyomo.base.log_config
from coopr.pyomo.base.numvalue import *
from coopr.pyomo.base.expr import *
from coopr.pyomo.base.expression import *
from coopr.pyomo.base.intrinsic_functions import *
from coopr.pyomo.base.label import *
from coopr.pyomo.base.plugin import *
from coopr.pyomo.base.PyomoModelData import *
from coopr.pyomo.base.DataPortal import *
from coopr.pyomo.base.symbol_map import *
#
# Components
#
from coopr.pyomo.base.component import *
from coopr.pyomo.base.sets import *
from coopr.pyomo.base.param import *
from coopr.pyomo.base.var import *
from coopr.pyomo.base.constraint import *
from coopr.pyomo.base.objective import *
from coopr.pyomo.base.connector import *
from coopr.pyomo.base.sos import *
from coopr.pyomo.base.piecewise import *
from coopr.pyomo.base.suffix import *
from coopr.pyomo.base.external import *
#
from coopr.pyomo.base.set_types import *
from coopr.pyomo.base.misc import *
from coopr.pyomo.base.block import *
from coopr.pyomo.base.PyomoModel import *
#
import coopr.pyomo.base.pyomo
#
from coopr.pyomo.base.util import *
from coopr.pyomo.base.rangeset import *

from coopr.pyomo.base.instance2dat import *

#
# This is a hack to strip out modules, which shouldn't have been included in these imports
#
import types
_locals = locals()
__all__ = [__name for __name in _locals.keys() if (not __name.startswith('_') and not isinstance(_locals[__name],types.ModuleType)) or __name == '_' ]
__all__.append('pyomo')
